## yapCAD boolean operations and surface creation geometry example

from yapcad.geom import *
from yapcad.poly import *
from yapcad.combine import *
from yapcad.geom_util import *
from yapcad.geom3d import *

import examples.example8 as example8

docGeomList=[]

def makeJigsawPiece(width=100,height=50,diam=30,frac=0.80,upper=False,poly=False):
    rect = RoundRect(width,height,5.0)
    rect.translate(point(0,-height/2))
    circ1 = Circle(point(-width/4,(diam/2)*frac),diam/2)
    #circ2 = Circle(point(width/4,-(diam/2)*frac),diam/2)
    circ2 = RoundRect(diam,diam,20.0,center=point(width/4,0))

    docGeomList.append(circ1.geom)
    docGeomList.append(circ2.geom)
    bln = Boolean('union',[circ1,rect])
    bln = Boolean('difference',[bln,circ2])
    docGeomList.append(bln.geom)
    if poly:
        return bln
    return bln.geom

def drawLegend(d):
    ## Put some documentary text on the drawing
    d.layer = 'DOCUMENTATION'

    att = {'style': 'OpenSans-Bold',
           'height': 2.5}
    
    d.draw_text("yapCAD", point(5,15), attr = att)

    d.draw_text("example12.py", point(5,11), attr=att)
    d.draw_text("complex boolean operations on Polygon() instances",
                point(5,7),attr=att)
    d.layer = False # back to default layer


def geometry():

    logopoly = example8.flower(maxRadius=30,
                               returnPoly=True)
    jigsaw=makeJigsawPiece(poly=True)
    bigc= Circle(point(0,0,0),20)
    nb = Boolean('union',[logopoly,bigc])
    nb.translate(point(0,-50))
    docGeomList.append(nb.geom)

    nb = Boolean('difference',[jigsaw,nb])
    
    gl = nb.geom
    gl2 = []

    for i in range(100):
        u = i/99.0
        rad = 1.0 + u*2
        
        p = nb.sample(u)
        c = arc(p,rad)
        gl2.append(c)
        

    return gl,gl2


def testAndDraw(dd):
    geom,geom2=geometry()

    if not renderOgl:
        dd.layer = 'DRILLS'
        dd.linecolor = False

    else:
        dd.linecolor = 'aqua'
    dd.draw(geom2)
    if renderOgl:
        dd.linecolor = 'red'
        dd.draw(translate(geom,point(0,0,5)))
    else:
        dd.draw(geom)

    if not renderOgl:
        dd.layer = 'DOCUMENTATION'
        dd.linecolor = False
    else:
        dd.linecolor = 'red'
    dd.draw(docGeomList)

    if not renderOgl:
        dd.layer = 'PATHS'
        
    dd.linecolor = 'white'
    if renderOgl:
        ply = geomlist2poly(translate(geom,point(0,0,-0.1)))
        surf,bnd = poly2surfaceXY(ply)
        dd.draw_surface(surf)
    
    geom = translate(geom,point(0,0,0.1))
    dd.draw(geom)

    dd.display()
    

if __name__ == "__main__":
    import sys
    renderOgl = False
    filename="example12-out"
    print("example12.py -- yapCAD boolean operations demonstration")

    oglarg= ("pyglet","opengl","OpenGL")
    dxfarg= ("ezdxf","dxf","DXF")
    if len(sys.argv) > 1:
        if sys.argv[1] in oglarg:
            renderOgl=True
        elif sys.argv[1] in dxfarg:
            renderOgl=False
        else:
            print("This is an example of a complex boolean operation on Polygon() and")
            print("Boolean() instances. When rendering with OpenGL, we also create a")
            print("two-dimensional surface from the resulting figure.")
            print("syntax: $ python3 {} <rendertype> [filename]".format(sys.argv[0]))
            print("    where <rendertype> is one of {} for OpenGL".format(oglarg))
            print("    or one of {} for DXF".format(dxfarg))
            print("    For DXF, you can optionally specify [filename].dxf as the output file")
            quit()
    if len(sys.argv) > 2 and renderOgl==False:
        filename = sys.argv[2]+".dxf"
    requestedOgl = renderOgl
    dd = None
    if renderOgl:
        print("OpenGL rendering selected")
        try:
            from yapcad.pyglet_drawable import *
            dd=pygletDraw()
            dd.cameradist=170.0
        except RuntimeError as err:
            print("\nSkipping OpenGL rendering: {}".format(err))
            renderOgl = False
        except Exception as err:
            print("\nSkipping OpenGL rendering due to unexpected error: {}".format(err))
            renderOgl = False

    if not renderOgl or dd is None:
        if requestedOgl and dd is None:
            print("Falling back to DXF rendering")
        else:
            print("DXF rendering selected")
        from yapcad.ezdxf_drawable import *
        #set up DXF rendering
        dd=ezdxfDraw()
        dd.filename = filename
        drawLegend(dd)
        renderOgl = False
    print("rendering...")
    
    testAndDraw(dd)
    print("done")



## yapCAD poly() intersecton and drawing examples

from yapcad.geom import *
from yapcad.poly import *
import examples.example10 as example10
import random

def drawLegend(d):
    ## Put some documentary text on the drawing
    d.layer = 'DOCUMENTATION'

    att = {'style': 'OpenSans-Bold',
           'height': 2.5}
    
    d.draw_text("yapCAD", point(5,15), attr = att)

    d.draw_text("example8.py", point(5,11), attr=att)
    d.draw_text("Polygon() flowers, mirrored geometry",
                point(5,7),attr=att)
    d.layer = False # back to default layer


def flower(center = point(0,0),
           petals = 10,
           minDiam=5.0,
           maxDiam=15,
           minRadius=20,
           maxRadius=40,
           insideRad = 5, returnPoly = False):

    glist= []
    for i in range(petals):
        angle = i*360/petals
        anrad = angle*pi2/360.0

        an2 = (((i+0.5)/petals)%1.0)*360
        an2rad = an2*pi2/360.0
    
        radius = (maxRadius-minRadius)*random.random()+minRadius
        diam = (maxDiam-minDiam)*random.random()+minDiam
        pnt = add(point(cos(anrad)*radius,sin(anrad)*radius),
                  center)
        pnt2 = add(point(cos(an2rad)*insideRad,sin(an2rad)*insideRad),
                   center)
        a = arc(pnt,diam/2)
        glist.append(a)
        glist.append(pnt2)

    p = Polygon(glist)
    #p.makeoutline()
    if returnPoly:
        return p
    else:
        return p.geom

def mirrorArray(pnt=point(-45,45)):
    flwr = flower(pnt,returnPoly=True)
    glist = flwr.geom
    bb = bbox(glist)
    flwr2 = deepcopy(flwr)
    flwr2.grow(1.0)
    glist.append(flwr2.geom)

    ranp = example10.randomPoints(bb,500)

    for p in ranp:
        if flwr.isinsideXY(p):
            glist.append(arc(p,0.4))
                         
    ply = [point(bb[0]), point(bb[1][0],bb[0][1]),
           point(bb[1]), point(bb[0][0],bb[1][1])]
    glist = glist + ply
    glist = glist + mirror(glist,'yz')
    glist = glist + mirror(glist,'xz')
    return glist

if __name__ == "__main__":

    from yapcad.ezdxf_drawable import *
    #set up DXF rendering

    print("example8.py -- yapCAD computational geometry and DXF drawing example")
    print("""
This is a demo of the Polygon() and mirror capabilities of yapCAD,
making and then mirroring random "flowers."
    """)
    dd=ezdxfDraw()
    filename="example8-out"
    print("Output file name is {}.dxf".format(filename))
    dd.filename = filename

    glist = mirrorArray()

    def mydrawer(gl):
        g1=[]
        g2=[]
        for g in gl:
            if iscircle(g):
                g1.append(g)
            else:
                g2.append(g)
        dd.linecolor = 'white'
        dd.draw(g2)
        dd.linecolor = 'aqua'
        dd.draw(g1)

    drawLegend(dd)
    mydrawer(glist)
    dd.display()



"""Common geometric helpers shared across exporters and validators."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Sequence, Tuple

from yapcad.geom import cross, epsilon, mag

Vec3 = Tuple[float, float, float]


@dataclass(frozen=True)
class Triangle:
    """Immutable triangle representation in XYZ space."""

    normal: Vec3
    v0: Vec3
    v1: Vec3
    v2: Vec3


def to_vec3(point_like: Sequence[float]) -> Vec3:
    """Return the XYZ components of a yapCAD point/vector as a tuple."""

    if len(point_like) < 3:
        raise ValueError("value must have at least three components")
    return float(point_like[0]), float(point_like[1]), float(point_like[2])


def to_point(vec: Sequence[float]) -> Tuple[float, float, float, float]:
    """Lift an XYZ tuple into yapCAD homogeneous point form."""

    if len(vec) < 3:
        raise ValueError("vector must have three components")
    return float(vec[0]), float(vec[1]), float(vec[2]), 1.0


def to_vector(vec: Sequence[float]) -> Tuple[float, float, float, float]:
    """Lift an XYZ tuple into yapCAD homogeneous vector form (w=0)."""

    if len(vec) < 3:
        raise ValueError("vector must have three components")
    return float(vec[0]), float(vec[1]), float(vec[2]), 0.0


def triangle_normal(v0: Vec3, v1: Vec3, v2: Vec3) -> Vec3 | None:
    """Return the unit normal of a triangle or ``None`` if degenerate."""

    ax, ay, az = v1[0] - v0[0], v1[1] - v0[1], v1[2] - v0[2]
    bx, by, bz = v2[0] - v0[0], v2[1] - v0[1], v2[2] - v0[2]
    n = cross([ax, ay, az, 1.0], [bx, by, bz, 1.0])
    length = mag(n)
    if length <= epsilon:
        return None
    return (n[0] / length, n[1] / length, n[2] / length)


def triangle_area(v0: Vec3, v1: Vec3, v2: Vec3) -> float:
    """Return the area of a triangle."""

    ax, ay, az = v1[0] - v0[0], v1[1] - v0[1], v1[2] - v0[2]
    bx, by, bz = v2[0] - v0[0], v2[1] - v0[1], v2[2] - v0[2]
    n = cross([ax, ay, az, 1.0], [bx, by, bz, 1.0])
    return 0.5 * mag(n)


def triangle_is_degenerate(v0: Vec3, v1: Vec3, v2: Vec3, tol: float = epsilon) -> bool:
    """Return ``True`` if a triangle collapses under the given tolerance."""

    return triangle_area(v0, v1, v2) <= tol


def orient_triangle(v0: Vec3, v1: Vec3, v2: Vec3, preferred_normal: Vec3) -> Tuple[Vec3, Vec3, Vec3]:
    """Ensure triangle winding aligns with ``preferred_normal``."""

    current = triangle_normal(v0, v1, v2)
    if current is None:
        return v0, v1, v2
    dot = current[0] * preferred_normal[0] + current[1] * preferred_normal[1] + current[2] * preferred_normal[2]
    if dot < 0:
        return v0, v2, v1
    return v0, v1, v2


def triangle_centroid(v0: Vec3, v1: Vec3, v2: Vec3) -> Vec3:
    """Return the centroid of a triangle."""

    return (
        (v0[0] + v1[0] + v2[0]) / 3.0,
        (v0[1] + v1[1] + v2[1]) / 3.0,
        (v0[2] + v1[2] + v2[2]) / 3.0,
    )


def triangles_from_mesh(mesh: Iterable[Tuple[Vec3, Vec3, Vec3, Vec3]]) -> Iterable[Triangle]:
    """Convert ``mesh_view`` output into ``Triangle`` instances."""

    for normal, v0, v1, v2 in mesh:
        yield Triangle(normal=normal, v0=v0, v1=v1, v2=v2)


__all__ = [
    "Triangle",
    "Vec3",
    "to_vec3",
    "to_point",
    "to_vector",
    "triangle_normal",
    "triangle_area",
    "triangle_is_degenerate",
    "orient_triangle",
    "triangle_centroid",
    "triangles_from_mesh",
]



## yapCAD 3D geometry example

from yapcad.geom import *
from yapcad.geom3d import *
from yapcad.geom3d_util import *

import math


if __name__ == "__main__":
    import sys
    try:
        from yapcad.pyglet_drawable import *
        dd=pygletDraw()
    except RuntimeError as err:
        print("\nSkipping OpenGL rendering: {}".format(err))
        sys.exit(0)
    except Exception as err:
        print("\nSkipping OpenGL rendering due to unexpected error: {}".format(err))
        sys.exit(1)
    print("example9.py -- yapCAD 3D geometry demonstration")
    print("""
This is a demonstration of the creation and rendering of
three-dimensional geometry using yapCAD and the pyglet openGL draing
engine.

In this example we create an icosohedron centered at the orign and
tesellate it spherically""")

    dd.linecolor = 'white'

    # rotate the camera down
    dd.ry = -80.0

    # make a spherical surface
    sphereobj = sphere(50.0,point(0,0,0),3)

    # make a rectangular prism
    pris = prism(50,30,20)

    # make a cone

    cone = conic(20,4,20)

    # set up some parameters for animation

    # these timers could be consolidated
    time1 = 0
    time2 = 0

    # paths for objects to move on
    circle1 = rotate(geomlist2poly([ arc(point(-30,0,0),25) ]),
                     45,axis=point(1.0/sqrt(2),1.0/sqrt(2),0))
    circle2 = arc(point(30,0,0),50)
    
    def animateSphere1(dt):
        global time1
        time1 += dt
        obj = dd.objectdict['sphere1']

        pos = sample(circle1,(time1/10)%1.0)
        obj.x = pos[0]
        obj.y = pos[1]
        obj.z = pos[2]
        obj.rz = time1*10.0

    def animateSphere2(dt):
        global time2
        time2 += dt
        obj = dd.objectdict['sphere2']

        pos = sample(circle2,(time2/10+0.5)%1.0)
        obj.x = pos[0]
        obj.y = pos[1]
        obj.z = pos[2]
        obj.ry = -time2*10.0

    ## create some aminatable, renderable objects

    dd.make_object('sphere0',lighting=False,linecolor='white',
                   position=point(0,0,0))
    
    dd.make_object('sphere1',lighting=True,material='jade',
                   linecolor='green',
                   animate=animateSphere1)
    
    dd.make_object('sphere2',lighting=True,material='polished bronze',
                   linecolor='yellow',
                   animate=animateSphere2)

    dd.draw(sphereobj,name='sphere0')
    dd.draw(sphereobj,name='sphere1')
    #dd.draw(sphereobj,name='sphere2')
    #dd.draw(pris,name='sphere2')
    dd.draw(cone,name='sphere2')


    # for the stationary sphere, draw some additional points
    dd.linecolor = 'aqua'
    dd.polystyle = 'points'
    dd.pointstyle = 'xo'
    
    spheresurf = sphereobj[1][0]
    dd.draw(spheresurf[1]) # the verticies of the sphere
    dd.display()



import copy
import pytest

from yapcad.geom import point
from yapcad.geom3d import (
    reversesurface,
    solid_boolean,
    solid_contains_point,
    solids_intersect,
    translatesolid,
    _iter_triangles_from_solid,
    tri2p0n,
)
from yapcad.geom3d_util import prism, sphere, conic, tube


def _offset_point(base, normal, scale):
    return point(base[0] + normal[0] * scale,
                 base[1] + normal[1] * scale,
                 base[2] + normal[2] * scale)


def _assert_normals_outward(sld, *, label='solid', step=1e-3):
    triangles = list(_iter_triangles_from_solid(sld))
    assert triangles, f"{label} produced no faces"

    max_checks = min(32, len(triangles))
    stride = max(1, len(triangles) // max_checks)
    checked = 0

    for idx, tri in enumerate(triangles):
        if idx % stride != 0:
            continue
        try:
            center, normal = tri2p0n(tri)
        except ValueError:
            continue

        scale = step
        detected = False
        flipped = False
        for _ in range(6):
            inside_pt = _offset_point(center, normal, -scale)
            outside_pt = _offset_point(center, normal, scale)
            inside = solid_contains_point(sld, inside_pt)
            outside = solid_contains_point(sld, outside_pt)
            if inside and not outside:
                detected = True
                break
            if outside and not inside:
                flipped = True
                break
            scale *= 2.0
        if flipped:
            raise AssertionError(
                f"normal orientation check failed for {label} at center {center[:3]}"
            )
        if detected:
            checked += 1

    assert checked > 0, f"no triangles sampled for {label}"


def _translate(sld, delta):
    return translatesolid(sld, point(delta[0], delta[1], delta[2]))


def _make_cavity_solid():
    outer = prism(2, 2, 2)
    inner = prism(1, 1, 1)
    cavity_surfaces = outer[1] + [reversesurface(s) for s in inner[1]]
    return ['solid', cavity_surfaces, [], []]


def test_solid_contains_point_basic():
    cube = prism(2, 2, 2)
    assert solid_contains_point(cube, point(0.1, 0.1, 0.1))
    assert solid_contains_point(cube, point(1.0, 0.0, 0.0))
    assert not solid_contains_point(cube, point(2.5, 0.0, 0.0))


def test_solid_contains_point_cavity():
    cavity = _make_cavity_solid()
    assert not solid_contains_point(cavity, point(0.1, 0.0, 0.0))
    assert solid_contains_point(cavity, point(0.9, 0.0, 0.0))


@pytest.mark.slow
def test_solid_boolean_cubes():
    a = prism(2, 2, 2)
    b = _translate(prism(2, 2, 2), (0.75, 0.0, 0.0))

    union = solid_boolean(a, b, 'union')
    assert solid_contains_point(union, point(-0.9, 0.0, 0.0))
    assert solid_contains_point(union, point(1.6, 0.0, 0.0))

    intersection = solid_boolean(a, b, 'intersection')
    assert solid_contains_point(intersection, point(0.6, 0.0, 0.0))
    assert not solid_contains_point(intersection, point(-0.9, 0.0, 0.0))

    difference = solid_boolean(a, b, 'difference')
    assert solid_contains_point(difference, point(-0.9, 0.0, 0.0))
    assert not solid_contains_point(difference, point(0.6, 0.0, 0.0))


@pytest.mark.slow
def test_solid_boolean_spheres():
    a = sphere(2.0)
    b = copy.deepcopy(a)
    b = translatesolid(b, point(1.0, 0.0, 0.0))

    union = solid_boolean(a, b, 'union')
    assert solid_contains_point(union, point(-0.8, 0.0, 0.0))
    assert solid_contains_point(union, point(1.8, 0.0, 0.0))

    intersection = solid_boolean(a, b, 'intersection')
    assert solid_contains_point(intersection, point(0.5, 0.0, 0.0))
    assert not solid_contains_point(intersection, point(-1.8, 0.0, 0.0))

    difference = solid_boolean(a, b, 'difference')
    assert solid_contains_point(difference, point(-0.8, 0.0, 0.0))
    assert not solid_contains_point(difference, point(0.5, 0.0, 0.0))


@pytest.mark.slow
def test_solids_intersect_detection():
    base = prism(2, 2, 2)
    shifted = _translate(prism(2, 2, 2), (0.5, 0.0, 0.0))
    assert solids_intersect(base, shifted)

    far = _translate(prism(2, 2, 2), (10.0, 0.0, 0.0))
    assert not solids_intersect(base, far)


@pytest.mark.slow
def test_solid_boolean_sphere_normals():
    a = sphere(2.0)
    b = copy.deepcopy(a)
    b = translatesolid(b, point(1.0, 0.0, 0.0))

    union = solid_boolean(a, b, 'union')
    _assert_normals_outward(union, label='sphere union')

    difference = solid_boolean(a, b, 'difference')
    _assert_normals_outward(difference, label='sphere difference')


@pytest.mark.slow
def test_solid_boolean_box_minus_cylinder():
    box = prism(2, 2, 2)
    drill = conic(0.6, 0.6, 3.0, center=point(0.0, 0.0, -1.5))

    result = solid_boolean(box, drill, 'difference')

    assert not solid_contains_point(result, point(0.0, 0.0, 0.0))
    assert solid_contains_point(result, point(0.9, 0.0, 0.0))
    _assert_normals_outward(result, label='box minus cylinder')


@pytest.mark.slow
def test_solid_boolean_tube_side_hole():
    shell = tube(outer_diameter=3.0, wall_thickness=0.5, length=4.0,
                 base_point=point(0.0, 0.0, -2.0))
    drill = conic(0.6, 0.6, 4.2, center=point(1.2, 0.0, -2.1))

    result = solid_boolean(shell, drill, 'difference')

    assert not solid_contains_point(result, point(1.2, 0.0, 0.0))
    assert solid_contains_point(result, point(0.0, 1.3, 0.0))
    _assert_normals_outward(result, label='tube side hole')



## utilty functions companion to yapcad.geom
## Born on 26 December, 2020
## Copyright (c) 2020 Richard DeVaul

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from yapcad.geom import *
from yapcad.spline import sample_catmullrom, sample_nurbs
import math
import random

""" 
Less-essential yapCAD utility functions to support the creation and
operations on yapcad.geom figures.

"""


def _reverse_element(element):
    """Return a geometry element with its orientation reversed."""

    return reverseGeomList([element])[0]


def _poly_signed_area(points):
    """Compute signed area of a closed polygon represented by ``points``."""

    if not points:
        return 0.0
    total = 0.0
    for i in range(1, len(points)):
        x1, y1 = points[i-1][0], points[i-1][1]
        x2, y2 = points[i][0], points[i][1]
        total += (x1 * y2) - (x2 * y1)
    return 0.5 * total


def _stitch_geomlist(gl, tol):
    """Group line/arc primitives from ``gl`` into contiguous loops."""

    segments = [g for g in gl if isline(g) or isarc(g)]
    others = [g for g in gl if not (isline(g) or isarc(g))]

    if not segments:
        return [], others

    nodes = []

    def node_id(p):
        for idx, existing in enumerate(nodes):
            if dist(existing, p) <= tol:
                return idx
        nodes.append(point(p))
        return len(nodes) - 1

    edges = []
    adjacency = {}

    for idx, seg in enumerate(segments):
        start = sample(seg, 0.0)
        end = sample(seg, 1.0)
        s_id = node_id(start)
        e_id = node_id(end)
        edges.append({'geom': seg, 'start': s_id, 'end': e_id, 'used': False})
        adjacency.setdefault(s_id, []).append(idx)
        adjacency.setdefault(e_id, []).append(idx)

    loops = []

    for idx in range(len(edges)):
        edge = edges[idx]
        if edge['used']:
            continue
        loop = []
        current_idx = idx
        start_node = edge['start']
        prev_node = start_node
        while True:
            edge = edges[current_idx]
            if edge['used']:
                break
            if edge['start'] != prev_node:
                if dist(nodes[edge['end']], nodes[prev_node]) <= tol:
                    edge['geom'] = _reverse_element(edge['geom'])
                    edge['start'], edge['end'] = edge['end'], edge['start']
                elif dist(nodes[edge['start']], nodes[prev_node]) > tol:
                    break
            loop.append(edge['geom'])
            edge['used'] = True
            prev_node = edge['end']
            if len(loop) > 1 and dist(nodes[prev_node], nodes[start_node]) <= tol:
                break
            next_idx = None
            for cand_idx in adjacency.get(prev_node, []):
                if edges[cand_idx]['used']:
                    continue
                next_idx = cand_idx
                break
            if next_idx is None:
                break
            current_idx = next_idx
        if loop:
            loops.append(loop)

    return loops, others


def _finalize_poly_points(points, snap_tol):
    if not points:
        return []

    ply = [point(points[0])]
    for p in points[1:]:
        if dist(p, ply[-1]) > epsilon:
            ply.append(point(p))

    if dist(ply[0], ply[-1]) > epsilon:
        if dist(ply[0], ply[-1]) <= snap_tol:
            ply[-1] = point(ply[0])
        else:
            ply.append(point(ply[0]))
    else:
        ply[-1] = point(ply[0])

    return ply


def _convert_geom_sequence(seq, minang, minlen, snap_tol):
    ply = []
    lastpoint = None
    nested_holes = []

    def _snap(p, last):
        pp = point(p)
        if last is not None and dist(pp, last) <= snap_tol:
            return point(last)
        return pp

    def addpoint(p, last, force=False):
        candidate = _snap(p, last)
        if force or last is None or dist(candidate, last) > minlen:
            last = point(candidate)
            ply.append(last)
        return last

    for element in seq:
        if ispoint(element):
            continue
        elif isline(element):
            lastpoint = addpoint(element[0], lastpoint, force=True)
            lastpoint = addpoint(element[1], lastpoint, force=True)
        elif isarc(element):
            firstpoint = None
            for ang in range(0, 360, max(1, round(minang))):
                p = polarSampleArc(element, float(ang), inside=True)
                if not p:
                    break
                if firstpoint is None:
                    firstpoint = p
                lastpoint = addpoint(p, lastpoint)
            if iscircle(element):
                lastpoint = addpoint(firstpoint, lastpoint)
            else:
                endp = sample(element, 1.0)
                if endp:
                    lastpoint = addpoint(endp, lastpoint, force=True)
        elif ispoly(element):
            for p in element:
                lastpoint = addpoint(p, lastpoint, force=True)
        elif isgeomlist(element):
            sub_outer, sub_holes = geomlist2poly_with_holes(element, minang, minlen, checkcont=False)
            if sub_outer:
                for p in sub_outer:
                    lastpoint = addpoint(p, lastpoint, force=True)
            nested_holes.extend(sub_holes)
        else:
            raise ValueError(f'bad object in list passed to geomlist2poly: {element}')

    if not ply:
        return [], nested_holes

    cleaned = [ply[0]]
    for p in ply[1:]:
        if dist(p, cleaned[-1]) > epsilon:
            cleaned.append(point(p))

    if dist(cleaned[0], cleaned[-1]) > epsilon:
        if dist(cleaned[0], cleaned[-1]) <= snap_tol:
            cleaned[-1] = point(cleaned[0])
        else:
            cleaned.append(point(cleaned[0]))
    else:
        cleaned[-1] = point(cleaned[0])

    return cleaned, nested_holes


def _points_to_segments(points, *, closed=False):
    segments = []
    if not points:
        return segments
    prev = point(points[0])
    for pt in points[1:]:
        curr = point(pt)
        if dist(prev, curr) > epsilon:
            segments.append(line(prev, curr))
        prev = curr
    if closed:
        start = point(points[0])
        if dist(prev, start) > epsilon:
            segments.append(line(prev, start))
    return segments


def _spline_segments_from_curve(curve, minlen):
    if iscatmullrom(curve):
        meta = curve[2]
        default_segments = max(8, int(round(1.0 / max(minlen, epsilon))))
        segs = int(meta.get('segments_per_span', default_segments))
        samples = sample_catmullrom(curve, segments_per_span=max(1, segs))
        closed = bool(meta.get('closed', False))
        return _points_to_segments(samples, closed=closed)
    if isnurbs(curve):
        meta = curve[2]
        default = max(32, len(curve[1]) * 8)
        count = int(meta.get('samples', default))
        samples = sample_nurbs(curve, samples=max(4, count))
        return _points_to_segments(samples, closed=False)
    return []

def _collect_candidate_polygons(gl, minang, minlen, snap_tol):
    """Return raw polygon loops discovered within ``gl``."""

    loops = []
    segments = []

    for element in gl:
        if ispoint(element):
            continue
        if isline(element) or isarc(element):
            segments.append(element)
            continue
        if iscatmullrom(element) or isnurbs(element):
            segments.extend(_spline_segments_from_curve(element, minlen))
            continue
        if ispoly(element):
            loops.append(_finalize_poly_points(element, snap_tol))
            continue
        if isgeomlist(element):
            loops.extend(_collect_candidate_polygons(element, minang, minlen, snap_tol))
            continue
        raise ValueError(f'bad object in list passed to geomlist2poly: {element}')

    if segments:
        loops_seq, leftovers = _stitch_geomlist(segments, snap_tol)
        for loop in loops_seq:
            poly, holes = _convert_geom_sequence(loop, minang, minlen, snap_tol)
            if poly:
                loops.append(poly)
            loops.extend(holes)
        for leftover in leftovers:
            if ispoly(leftover):
                loops.append(_finalize_poly_points(leftover, snap_tol))
            elif isgeomlist(leftover):
                loops.extend(_collect_candidate_polygons(leftover, minang, minlen, snap_tol))
            elif ispoint(leftover):
                continue
            elif isline(leftover) or isarc(leftover):
                poly, holes = _convert_geom_sequence([leftover], minang, minlen, snap_tol)
                if poly:
                    loops.append(poly)
                loops.extend(holes)
            elif iscatmullrom(leftover) or isnurbs(leftover):
                segments = _spline_segments_from_curve(leftover, minlen)
                if segments:
                    poly, holes = _convert_geom_sequence(segments, minang, minlen, snap_tol)
                    if poly:
                        loops.append(poly)
                    loops.extend(holes)
            else:
                raise ValueError(f'bad object in list passed to geomlist2poly: {leftover}')

    return loops


def _orient_loop(points, want_positive=True):
    if not points:
        return []
    core = points[:-1] if len(points) > 1 else points
    if len(core) < 3:
        return [point(p) for p in points]
    pts = [point(p) for p in core]
    area = _poly_signed_area(points)
    should_reverse = (want_positive and area < 0) or (not want_positive and area > 0)
    if should_reverse:
        pts.reverse()
    pts.append(point(pts[0]))
    return pts


def _assemble_components(loop_points):
    if not loop_points:
        return []

    nodes = []
    for pts in loop_points:
        if len(pts) < 3:
            continue
        area = _poly_signed_area(pts)
        if abs(area) <= epsilon:
            continue
        nodes.append({
            'points': pts,
            'area': area,
            'abs_area': abs(area),
            'centroid': _polygon_centroid_xy(pts),
            'parent': None,
            'children': [],
        })

    if not nodes:
        return []

    for idx, node in enumerate(nodes):
        parent_idx = None
        for jdx, candidate in enumerate(nodes):
            if idx == jdx:
                continue
            if candidate['abs_area'] <= node['abs_area'] + epsilon:
                continue
            if _point_in_polygon_xy(candidate['points'], node['centroid']):
                if parent_idx is None or candidate['abs_area'] < nodes[parent_idx]['abs_area']:
                    parent_idx = jdx
        node['parent'] = parent_idx
        if parent_idx is not None:
            nodes[parent_idx]['children'].append(idx)

    components = []

    def visit(node_idx):
        node = nodes[node_idx]
        outer = _orient_loop(node['points'], want_positive=True)
        holes = []
        extras = []
        for child_idx in node['children']:
            child = nodes[child_idx]
            holes.append(_orient_loop(child['points'], want_positive=False))
            for grand_idx in child['children']:
                extras.extend(visit(grand_idx))
        extras.insert(0, (outer, holes))
        return extras

    for idx, node in enumerate(nodes):
        if node['parent'] is None:
            components.extend(visit(idx))

    return components


def geomlist2poly_components(gl, minang=5.0, minlen=0.25, checkcont=False):
    """Return a list of ``(outer, holes)`` tuples extracted from ``gl``."""

    if checkcont and not iscontinuousgeomlist(gl):
        raise ValueError('non-continuous geometry list passed to geomlist2poly')

    snap_tol = max(epsilon * 10, minlen * 0.1)
    loops = _collect_candidate_polygons(gl, minang, minlen, snap_tol)
    return _assemble_components(loops)


def geomlist2poly_with_holes(gl, minang=5.0, minlen=0.25, checkcont=False):
    components = geomlist2poly_components(gl, minang, minlen, checkcont)
    if not components:
        return [], []
    return components[0]


def _polygon_centroid_xy(pts):
    if not pts:
        return (0.0, 0.0)
    sx = sy = 0.0
    count = 0
    for p in pts:
        sx += p[0]
        sy += p[1]
        count += 1
    if count == 0:
        return (0.0, 0.0)
    return (sx / count, sy / count)


def _point_in_polygon_xy(poly, pt):
    x, y = pt
    inside = False
    if not poly:
        return False
    j = len(poly) - 1
    for i in range(len(poly)):
        xi, yi = poly[i][0], poly[i][1]
        xj, yj = poly[j][0], poly[j][1]
        intersects = ((yi > y) != (yj > y)) and (
            x < (xj - xi) * (y - yi) / (yj - yi + 1e-16) + xi)
        if intersects:
            inside = not inside
        j = i
    return inside

def randomPoints(bbox,numpoints):
    """Given a 3D bounding box and a number of points to generate, 
    return a list of uniformly generated random points within the 
    bounding box"""
    
    points = []
    minx = bbox[0][0]
    maxx = bbox[1][0]
    miny = bbox[0][1]
    maxy = bbox[1][1]
    minz = bbox[0][2]
    maxz = bbox[1][2]
    rangex = maxx-minx
    rangey = maxy-miny
    rangez = maxz-minz
    for i in range(numpoints):
        points.append(point(random.random()*rangex+minx,
                            random.random()*rangey+miny,
                            random.random()*rangez+minz))
    return points

def randomCenterInBox(bbox,r):
    """given a bounding box and a radius, generate a random center point
    such that a circle with the specified radius and center point falls
    completely inside the box

    """
    
    minx = bbox[0][0]+r
    maxx = bbox[1][0]-r
    miny = bbox[0][1]+r
    maxy = bbox[1][1]-r
    minz = bbox[0][2]+r
    maxz = bbox[1][2]-r
    rangex = maxx-minx
    rangey = maxy-miny
    rangez = maxz-minz
    x = point(random.random()*rangex+minx,
              random.random()*rangey+miny,
              random.random()*rangez+minz)
    return x


def randomArc(bbox,minr=0.0,maxr=10.0,circle=False):
    """given a bounding box and a minimum and maximum radius, generate an
    arc with random center and radius that falls within the bounding
    box.  If ``circle==False``, use randomly generated start and end
    angles, otherwise generate only full circles

    """
    radr = maxr-minr
    r = random.random()*radr+minr
    start = 0
    end = 360
    if not circle:
        start = random.random()*360.0
        end = start + random.random()*360.0

    x = randomCenterInBox(bbox,r)
    return arc(x,r,start,end)

def randomPoly(bbox,numpoints=10,minr = 1.0,maxr = 10.0):
    """given a bounding box, a number of vertices, and a minimum and
    maximum radius, generate a simple polygon that completely lies
    inside the bounding box, whose vertices are evenly spaced by angle
    around a center point, with randomly chosen radius between
    ``minr`` and ``maxr``
    """

    angles = []
    rads = []
    ang = 0.0
    rr = maxr-minr
    for i in range(numpoints):
        a = random.random()
        angles.append(ang)
        ang = ang + a
        # print("a: ",a," ang: ",ang)
        rads.append(random.random()*rr+minr)

    sf = pi2/ang

    points = []
    x = randomCenterInBox(bbox,maxr)
    
    for i in range(numpoints):
        p = [cos(angles[i]*sf)*rads[i],
             sin(angles[i]*sf)*rads[i],0,1]
        points.append(add(p,x))

    return points + [ points[0] ]

def makeLineSpiral(center, turnRad, # radius after one full turn
                   turns, # number of turns
                   dstep = 10.0, # sampling resolution in degrees
                   minlen = 0.25): # minimum distance between points
    """given a center point, the increase in radius per turn, the number
    of turns, and the angular resolution of the approximation,
    generate a yapcqad.geom poly approximation of the spiral

    """

    # make a spiral of points
    spiral = []
    rstep = turnRad*dstep/360.0

    def addpoint(p,lastpoint):
        if not lastpoint or dist(p,lastpoint) > minlen:
            lastpoint = p
            spiral.append(p)
        return lastpoint

    lastpoint = None
    for i in range(round(360*turns/dstep)):
        ang = i * dstep*pi2/360.0
        r = i * rstep
        p = add(center,
                point(math.cos(ang)*r,math.sin(ang)*r))
        lastpoint = addpoint(p,lastpoint)
    return spiral

def makeArcSpiral(center, turnRad, # radius after one full turn
                  turns, # number of turns
                  dstep = 45): # sampling resolution in degrees
    """given a center point, the increase in radius per turn, the number
    of turns, and the angular resolution of the approximation,
    generate an approximation of the spiral using circular arcs as
    segments instead of straightlines. Return a yapcqad.geom geomlist
    approximation of the spiral

    """
    spPoints = makeLineSpiral(center,turnRad,turns,dstep)

    arcs=[]
    for i in range(1,len(spPoints)):
        p0 = spPoints[i-1]
        p1 = spPoints[i]
        direct = sub(p1,p0)
        orth = scale3(orthoXY(direct),-1.1)
        mid = scale3(add(p0,p1),0.5)
        cent = add(mid,orth)
        r= dist(cent,p0)
        r2 = dist(cent,p1)
        assert close(r,r2)
        circ = arc(cent,r)
        u1 = unsamplearc(circ,p0)
        u2 = unsamplearc(circ,p1)
        a = segmentarc(circ,u1,u2)
        arcs.append(a)
    return arcs

def polarSampleArc(c,ang,inside=True):
    """given an arc ``c`` , sample that arc at ``ang`` degrees from the
    start of the arc, proceeding clockwise.  If ``samplereverse`` is
    true, sample the arc at ``-ang`` degrees from the end, proceeding
    counterclockwise.  If ``c`` is not a complete circle and the
    specified ``ang`` falls outside the closed [start,end] interval
    and ``inside == True``, return ``False``, otherwise return the
    sampled value.

    """
    p = c[0]
    r = c[1][0]
    start=c[1][1]
    end=c[1][2]
    samplereverse = (c[1][3] == -2)
    circle = (start == 0 and end == 360)

    if not circle:
        start = start % 360.0
        end = end % 360.0
        if end < start:
            end += 360.0
        if inside and ang > (end - start):
            return False
    
    if samplereverse:
        start=end
        ang *= -1

    srad = (start+ang)*pi2/360.0
    q = scale3(vect(cos(srad),sin(srad)),r)
    return add(p,q)


def triarea(p1,p2,p3):
    """
    utility function to return the area of a triangle
    """
    v1=sub(p2,p1)
    v2=sub(p3,p1)
    cp = cross(v1,v2)
    return mag(cp)/2

def geomlist2poly(gl,minang=5.0,minlen=0.25,checkcont=False):
    outer, _ = geomlist2poly_with_holes(gl, minang, minlen, checkcont)
    return outer



def combineglist(g1,g2,operation):
    """function to perform set operations on geometry lists.

    ``g1`` and ``g2`` are closed figure geometry lists.
    ``operation`` is one of ``["union","intersection","difference"]``
    
    result is a potentially zero-length geometry list representing the
    result, which may or may not be simply connected, but should
    always be a closed figure.

    """

    def poly2lines(pol):
        l = []
        for i in range(1,len(pol)):
            l.append([pol[i-1],pol[i]])
        return l

    if ispoly(g1):
        g1 = poly2lines(g1)

    if ispoly(g2):
        g2 = poly2lines(g2)
        
    # if not (isclosedgeomlist(g1) and isclosedgeomlist(g2)):
    #     raise ValueError("bad arguments passed to combineglist")

    if not operation in ["union","intersection","difference"]:
        raise ValueError("bad operation specified for combineglist")
    
    bbox1 = bbox(g1)
    bbox2 = bbox(g2)
    try :
        inter = intersectXY(g1,g2,params=True)
    except ValueError:
        print("had a problem intersecting following geometries:")
        print("g1: ",g1)
        print("g2: ",g2)
        raise

    if inter != False and (inter[0] == False or inter[1] == False):
        raise ValueError('bad intersection list: ',inter)
    # Given parameter values that define two potential sub-arcs of
    # a figure, determine which sub-arc is valid by looking for
    # intersections between these parameter values. In the
    # condition where u2 is smaller than u1, it's possible that
    # this represents an arc in the counter-clockwise direction
    # between u2 and u1.  It also could represent a clockwise arc
    # that "wraps around" from u1 to (u2+1).  We will check for
    # values x1, x2 in ilist that are u2 < x1 < u1 and u1 < x2 <
    # (u2+1.0).  The existance of x1 or x2 rules out the
    # corresponding arc.  If neither x1 nor x2 exists, we bias
    # towards the counter-clockwise arc.

    def between(u1,u2,ilist):
        if len(ilist) < 2:
            raise ValueError('bad ilist')
            
        if u1 > u2 :
            if len(ilist) == 2:
                return u1,u2+1.0,False
            x1s = list(filter(lambda x: x > u2 and x < u1,ilist))
            x2s = list(filter(lambda x: x > u1 or x < u2,ilist))
            l1 = len(x1s)
            l2 = len(x2s)
            #print("u1: ",u1," u2: ",u2," ilist: ",ilist," x1s: ",x1s," x2s: ",x2s)
            #if l1 > 0 and l2 > 0:
            #    print('WARNING: intersections on both sides')

            if l1 > l2:
                # print("AA")
                if True or operation == 'union':
                    return u1,u2+1.0,False
                else:
                    return u2,u1,True
            else:
                # print("A-")
                return u2,u1,True

        else:
            if len(ilist) == 2:
                return u1,u2,False
            x1s = list(filter(lambda x: x > u1 and x < u2,ilist))
            x2s = list(filter(lambda x: x > u2 or x < u1,ilist))
            l1 = len(x1s)
            l2 = len(x2s)
            #print("u1: ",u1," u2: ",u2," ilist: ",ilist," x1s: ",x1s," x2s: ",x2s)
            #if l1 > 0 and l2 > 0:
            #    print('WARNING: intersections on both sides')
                
            if l1 > l2:
                # print("BB")

                if True or operation == 'union':
                    return u2,u1+1,False
                else:
                    return u1,u2,False
            else:
                # print("B-")
                return u1,u2,False
            

    ## utility to perform combination on one "segment"
    def cmbin(g1,g2,itr):
        if not isgeomlist(g1):
            g1 = [ g1 ]

        if not isgeomlist(g2):
            g2 = [ g2 ]
        
        g1s = itr[0][0]
        g1e = itr[0][1]
        g2s = itr[1][0]
        g2e = itr[1][1]
        
        seg = []
        ZLEN1=close(g1s,g1e)
        ZLEN2=close(g2s,g2e)

        g1reverse=False
        g2reverse=False
            
        if True or operation == 'difference':
            if g1e < g1s:
                g1e+=1.0
            #g1s,g1e,g1reverse = between(g1s,g1e,inter[0])
            #import pdb ; pdb.set_trace()
            g2s,g2e,g2reverse = between(g2s,g2e,inter[1])
            g2reverse = False
        else:
            if g1e < g1s:
                g1e+=1.0
            if g2e < g2s:
                g2e += 1.0

        #p1=sample(g1,((g1s+g1e)/2)%1.0)

        p1inside=0
        for i in range(5):
            u = (i+1)/6.0
            p = sample(g1,(u*g1e+(1.0-u)*g1s)%1.0)
            if isinsideXY(g2,p):
                p1inside=p1inside+1

        p2inside = 0
        for i in range(5):
            u = (i+1)/6.0
            p = sample(g2,(u*g2e+(1.0-u)*g2s)%1.0)
            if isinsideXY(g1,p):
                p2inside=p2inside+1
                
        #if p1inside > 0 and p2inside > 0:
        #    print("warning: inside test succeeded for both p1s and p2s: ",
        #          p1inside," ",p2inside)

        #if p1inside == 0 and p2inside == 0:
        #    print("warning: inside test failed for both p1s and p2s")
                
        #p2=sample(g2,((g2s+g2e)/2)%1.0)

        if ZLEN1 and ZLEN2:
            print ('both segments zero length')
            return []
        elif ZLEN2 and not ZLEN1:
            print ('zero length segment 2')
            if operation=='union':
                return segmentgeomlist(g1,g1s,g1e,closed=True)
            elif operation=='difference':
                return []
            else: #intersection
                return []
        elif ZLEN1 and not ZLEN2:
            print ('zero length segment 1')
            if operation=='union':
                if g2e < g2s:
                    g2e += 1.0
                return segmentgeomlist(g2,g2s,g2e,closed=True)
            else: # difference or intersection
                return []
            
        if operation == 'union':
            #if isinsideXY(g2,p1):
            if p1inside > p2inside:
                # if g2e < g2s:
                #     g2e += 1.0
                seg += segmentgeomlist(g2,g2s,g2e,closed=True)
            else:
                seg += segmentgeomlist(g1,g1s,g1e,closed=True)
        elif operation == 'intersection':
            #if isinsideXY(g2,p1):
            if p1inside > p2inside:
                seg += segmentgeomlist(g1,g1s,g1e,closed=True)
            else:
                # if g2e < g2s:
                #     g2e += 1.0
                #seg += segmentgeomlist(g2,g2s,g2e,reverse=g2reverse)
                seg += segmentgeomlist(g2,g2s,g2e,closed=True)
        elif operation == 'difference':
            s = []
            #if isinsideXY(g2,p1):
            if p1inside > p2inside:
                pass
            else:
                # print("rsort: ",vstr(inter))
                seg += segmentgeomlist(g1,g1s,g1e,closed=True)
                # print("g2s: ",g2s," g2e: ",g2e," g2reverse: ",g2reverse)
                s = segmentgeomlist(g2,g2s,g2e,closed=True,reverse=g2reverse)
                s = reverseGeomList(s)
                seg += s
            if len(inter[0]) > 2:
                pass
                #combineDebugGL.append(s)
                # print("seg: ",vstr(seg))

        return seg

    ## utility function to sort intersections into non-decreasing
    ## order
    def rsort(il):
        nl = []
        rl = []
        rr = []
        for i in range(len(il[0])):
            nl.append([il[0][i],il[1][i]])
        nl.sort(key=lambda x: x[0])
        for i in range(len(nl)):
            rl.append(nl[i][0])
            rr.append(nl[i][1])
        return [rl,rr]

    if inter == False: # disjoint, but bounding boxes might be
        # null or one could be inside the other
        if not bbox1 and bbox2: # g1 is empty, but g2 contains geometry
            if operation=='union':
                return g2
            else:
                return []
        if bbox1 and not bbox2: # g2 is empty, but g1 isn't
            if operation=='union' or operation=='difference':
                return g1
        if not bbox1 and not bbox2: # no geometry at all
            return []
        ## OK, no intersection but it is possible that one profile
        ## could be inside the other.  Do fast bounding box checks
        ## before doing intersection-based checking.
        if isinsidebbox(bbox1,bbox2[0]) and isinsidebbox(bbox1,bbox2[1]) \
           and isinsideXY(g1,sample(g2,0.0)): # g2 is inside g1
            ## g2 is inside g1
            if operation == 'union':
                return g1
            elif operation == 'intersection':
                return g2
            else: #difference, g2 is a hole in g1
                return [g1,g2]
        elif isinsidebbox(bbox2,bbox1[0]) and isinsidebbox(bbox2,bbox1[1]) \
             and isinsideXY(g2,sample(g1,0.0)): # g1 is inside g2
            ## g1 is indside g2
            if operation == 'union':
                return g2
            elif operation == 'intersection':
                return g1
            else: #difference, g2 has eaten g1
                return []
        else: # g1 and g2 are disjoint
            if operation == 'union':
                return [g1,g2]
            elif operation == 'difference':
                return g1
            else: #intersection
                return []
    if len(inter[0]) == 1 and len(inter[1]) == 1:
    ## single point of intersection:
        if operation == 'union':
            return [g1, g2]
        elif operation == 'difference':
            return g1
        else: #intersection
            return []
    ## There are two or more points of intersection.
    inter = rsort(inter)
    #print("rsort: ",vstr(inter))

    if len(inter[0]) %2 != 0:
        print("WARNING: odd number of intersections (",len(inter[0]),", unpredictable behavior may result")
    r = []
    for i in range(1,len(inter[0])+1):
        r += cmbin(g1,g2,[[inter[0][i-1],
                           inter[0][i%len(inter[0])]],
                          [inter[1][i-1],
                           inter[1][i%len(inter[1])]]])
    return r



"""Trimesh-backed boolean engine for yapCAD solids.

This engine is optional. It converts yapCAD solids to ``trimesh.Trimesh``
instances, dispatches boolean operations via :mod:`trimesh.boolean`, and
converts the resulting mesh back into a yapCAD solid.

Availability depends on both the ``trimesh`` package and at least one
boolean backend supported by ``trimesh`` (e.g. Blender, OpenSCAD, Cork).
"""

from __future__ import annotations

from typing import Iterable

import os
import numpy as np

try:
    import trimesh
except ImportError:  # pragma: no cover - optional dependency
    trimesh = None  # type: ignore[assignment]

from . import native as _native

ENGINE_NAME = "trimesh"


def engines_available() -> set[str]:
    """Return the set of trimesh boolean backends that are operational."""

    if trimesh is None:  # pragma: no cover - optional dependency
        return set()
    return set(trimesh.boolean.engines_available)


def is_available(backend: str | None = None) -> bool:
    """Check whether the engine can run (trimesh + backend present)."""

    available = engines_available()
    if not available:
        return False
    if backend is None:
        return True
    return backend in available


def _solid_to_mesh(sld) -> "trimesh.Trimesh":
    if trimesh is None:  # pragma: no cover - optional dependency
        raise RuntimeError("trimesh is not installed")

    triangles = list(_native._iter_triangles_from_solid(sld))
    if not triangles:
        return trimesh.Trimesh(vertices=np.zeros((0, 3)), faces=np.zeros((0, 3), dtype=np.int64), process=False)

    vertex_map: dict[tuple[float, float, float], int] = {}
    verts = []
    faces = []
    for tri in triangles:
        face_inds = []
        for pt in tri:
            key = (round(pt[0], 9), round(pt[1], 9), round(pt[2], 9))
            idx = vertex_map.get(key)
            if idx is None:
                idx = len(verts)
                vertex_map[key] = idx
                verts.append([pt[0], pt[1], pt[2]])
            face_inds.append(idx)
        faces.append(face_inds)

    verts_np = np.asarray(verts, dtype=float)
    faces_np = np.asarray(faces, dtype=np.int64)
    mesh = trimesh.Trimesh(vertices=verts_np, faces=faces_np, process=False)
    mesh.remove_unreferenced_vertices()
    if not mesh.is_watertight:
        mesh.merge_vertices()
        mesh.remove_duplicate_faces()
        mesh.remove_degenerate_faces()
        mesh.process(validate=True)
    try:
        mesh.fix_normals()
    except Exception:
        pass
    return mesh


def _mesh_to_solid(mesh: "trimesh.Trimesh", operation: str) -> list:
    triangles = np.asarray(mesh.triangles)
    if triangles.size == 0:
        return _native._geom3d().solid([], [], ['boolean', f'{ENGINE_NAME}:{operation}'])

    tri_points = []
    for tri in triangles:
        tri_points.append([
            _native.point(tri[0][0], tri[0][1], tri[0][2]),
            _native.point(tri[1][0], tri[1][1], tri[1][2]),
            _native.point(tri[2][0], tri[2][1], tri[2][2]),
        ])

    surface = _native._surface_from_triangles(tri_points)
    if surface is None:
        return _native._geom3d().solid([], [], ['boolean', f'{ENGINE_NAME}:{operation}'])
    return _native._geom3d().solid([surface], [], ['boolean', f'{ENGINE_NAME}:{operation}'])


def solid_boolean(a, b, operation: str, tol=_native._DEFAULT_RAY_TOL, *, stitch: bool = False,
                  backend: str | None = None):
    """Perform a boolean between ``a`` and ``b`` using trimesh."""

    if trimesh is None:  # pragma: no cover - optional dependency
        raise RuntimeError("trimesh is not installed; install trimesh to enable this engine")

    available = engines_available()
    if backend is not None and backend not in available:
        raise RuntimeError(
            f"trimesh backend '{backend}' is not available; install the appropriate binary (available: {available})"
        )
    if backend is None and not available:
        raise RuntimeError(
            "no trimesh boolean backends are available; install Blender, OpenSCAD, Cork, or another supported engine"
        )

    mesh_a = _solid_to_mesh(a)
    mesh_b = _solid_to_mesh(b)

    backup_cache = None
    if backend == 'blender':
        backup_cache = os.environ.get('ARCH_CACHE_LINE_SIZE')
        os.environ['ARCH_CACHE_LINE_SIZE'] = '64'

    op = operation.lower()
    try:
        if op == 'union':
            result = trimesh.boolean.union([mesh_a, mesh_b], engine=backend, check_volume=False)
        elif op == 'intersection':
            result = trimesh.boolean.intersection([mesh_a, mesh_b], engine=backend, check_volume=False)
        elif op == 'difference':
            result = trimesh.boolean.difference([mesh_a, mesh_b], engine=backend, check_volume=False)
        else:
            raise RuntimeError(f"unsupported boolean operation '{operation}' for trimesh engine")
    except Exception as exc:  # pragma: no cover - depends on external binaries
        raise RuntimeError(f"trimesh boolean operation failed: {exc}") from exc
    finally:
        if backend == 'blender':
            if backup_cache is None:
                os.environ.pop('ARCH_CACHE_LINE_SIZE', None)
            else:
                os.environ['ARCH_CACHE_LINE_SIZE'] = backup_cache

    if result is None or result.faces.size == 0:
        return _native._geom3d().solid([], [], ['boolean', f'{ENGINE_NAME}:{operation}'])

    return _mesh_to_solid(result, operation)


__all__ = ['ENGINE_NAME', 'is_available', 'solid_boolean', 'engines_available']



## simple yapCAD framework for openGL drawing using pyglet
## package
## Copyright (c) 2020 Richard W. DeVaul
## Copyright (c) 2020 yapCAD contributors
## All rights reserved

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math
import pyglet
import pyglet.font as font
import pyglet.gl as gl
import pyglet.graphics as graphics

from yapcad.geom import *
from yapcad.geom3d import *
import yapcad.drawable as drawable

## openGL utility functions
def vec(*args):
    return (gl.GLfloat * len(args))(*args)

## HTML document, instructions for user interaction
yapCAD_legend="""
<font face="OpenSans, Geneva, sans-serif" size="5" color="#f0f0ff"><b>yapCAD</b><br></font>
<font face="Verdana, Geneva, sans-serif" size="3" color="white"><b>3D drawing viewer</b></font><br>
<font face="Verdana, Geneva, sans-serif" size="1" color="white">
<br>
<b>up-arrow</b>: zoom in<br>
<b>down-arrow</b>: zoom out<br>
<b>l</b>: toggle lighting mode<br>
<b>p</b>: toggle ground plane<br>
<b>left-mouse drag</b>: rotate view<br>
<b>m</b>: toggle display of this message<br>
<b>ESC</b>: exit viewer<br>
</font>
"""

class Material():
    """
    Representation of an OpenGL-style material
    """

    def __repr__(self):
        return f"Material(ambient={self.ambient}, diffuse={self.diffuse}, specular={self.specular}, emission={self.emission}, shininess={self.shininess}, desc=\"{self.desc}\")"

    def __checkargs(self,v):
        if ( not (len(v) == 3 or len(v) == 4) or
             len(list(filter(lambda x: not isinstance(x,(int,float)),v))) > 0 or
             len(list(filter(lambda x: x < 0.0 or x > 1.0,v))) > 0):
            raise ValueError('bad arguments to property setter')
        if len(v) == 3:
            return [v[0],v[1],v[2],1.0]
        else:
            return v

    
    def __init__(self,**kwargs):
        self.__ambient = vec(0.2, 0.2, 0.2, 1.0)
        self.__diffuse = vec(0.5, 0.5, 0.5, 1.0)
        self.__specular = vec(0.5, 0.5, 0.5, 1.0)
        self.__emission = vec(0.0, 0.0, 0.0, 1.0)
        self.__shininess = 50
        self.__desc = ""
        if 'ambient' in kwargs:
            self.ambient = kwargs['ambient']
        if 'diffuse' in kwargs:
            self.diffuse = kwargs['diffuse']
        if 'specular' in kwargs:
            self.specular = kwargs['specular']
        if 'emission' in kwargs:
            self.emission = kwargs['emission']
        if 'shininess' in kwargs:
            self.shininess = kwargs['shininess']
        if 'desc' in kwargs:
            self.desc = kwargs['desc']

    @property
    def ambient(self):
        return list(self.__ambient)

    @ambient.setter
    def ambient(self,*v):
        if len(v) == 1:
            v = v[0]
        v = self.__checkargs(v)
        self.__ambient = vec(*v)
        
    @property
    def diffuse(self):
        return list(self.__diffuse)

    @diffuse.setter
    def diffuse(self,*v):
        if len(v) == 1:
            v = v[0]
        v = self.__checkargs(v)
        self.__diffuse = vec(*v)
        
    @property
    def specular(self):
        return list(self.__specular)

    @specular.setter
    def specular(self,*v):
        if len(v) == 1:
            v = v[0]
        v = self.__checkargs(v)
        self.__specular = vec(*v)

    @property
    def emission(self):
        return list(self.__emission)

    @emission.setter
    def emission(self,*v):
        if len(v) == 1:
            v = v[0]
        v = self.__checkargs(v)
        self.__emission = vec(*v)

    @property
    def shininess(self):
        return self.__shininess

    @shininess.setter
    def shininess(self,v):
        if not isinstance(v,(int,float)) or v < 0 or v > 128:
            raise ValueError('bad shininess')
        self.__shininess=v

    @property
    def desc(self):
        return self.__desc

    @desc.setter
    def desc(self,v):
        if not isinstance(v,str):
            raise ValueError('bad description')
        self.__desc = v

    @property
    def material(self):
        return pyglet.model.Material(self.desc,
                                     self.diffuse,
                                     self.ambient,
                                     self.specular,
                                     self.emission,
                                     self.shininess)

        

## global materials dictionary
materials = {}

materials['default'] = Material(ambient=[0.5, 0.3, 0.2, 1.0],
                                diffuse = [0.6, 0.5, 0.3, 1.0],
                                specular = [1.0, 0.8, 0.5, 1.0],
                                emission = [0.0, 0.0, 0.0, 1.0],
                                shininess = 20,
                                desc = "default material, dull gold")


materials['groundplane'] = Material(diffuse = [0.02, 0.02, 0.023, 1.0],
                                    ambient = [0.01, 0.01, 0.02, 1.0],
                                    specular = [0.05, 0.06, 0.08, 1.0],
                                    emission = [0.0, 0.0, 0.0, 1.0],
                                    shininess = 20,
                                    desc = "ground plane material")

## The following materials a from tables published at
## http://devernay.free.fr/cours/opengl/materials.html and
## http://www.it.hiof.no/~borres/j3d/explain/light/p-materials.html

materials['emerald'] = Material(ambient=[0.0215, 0.1745, 0.0215, 1.0],
                                diffuse=[0.07568, 0.61424, 0.07568, 1.0],
                                specular=[0.633, 0.727811, 0.633, 1.0],
                                emission = [0.0, 0.0, 0.0, 1.0],
                                shininess = 0.6*128,
                                desc= "emerald")

materials['jade'] = Material(ambient=[0.135, 0.2225, 0.1575, 1.0],
                             diffuse=[0.54, 0.89, 0.63, 1.0],
                             specular=[0.316228, 0.316228, 0.316228, 1.0],
                             emission = [0.0, 0.0, 0.0, 1.0],
                             shininess = 0.1*128)

materials['obsidian'] = Material(ambient=[0.05375, 0.05, 0.06625, 1.0],
                                 diffuse=[0.18275, 0.17, 0.22525, 1.0],
                                 specular=[0.332741, 0.328634, 0.346435, 1.0],
                                 emission = [0.0, 0.0, 0.0, 1.0],
                                 shininess = 0.3*128)

materials['pearl'] = Material(ambient=[0.25, 0.20725, 0.20725, 1.0],
                              diffuse=[0.9999, 0.829, 0.829, 1.0],
                              specular=[0.296648, 0.296648, 0.296648, 1.0],
                              emission = [0.0, 0.0, 0.0, 1.0],
                              shininess = 0.088*128)

materials['ruby'] = Material(ambient=[0.1745, 0.01175, 0.01175, 1.0],
                             diffuse=[0.61424, 0.04136, 0.04136, 1.0],
                             specular=[0.727811, 0.626959, 0.626959, 1.0],
                             emission = [0.0, 0.0, 0.0, 1.0],
                             shininess = 0.6*128)

materials['turquoise'] = Material(ambient=[0.1, 0.18725, 0.1745, 1.0],
                                  diffuse=[0.396, 0.74151, 0.69102, 1.0],
                                  specular=[0.297254, 0.30829, 0.306678, 1.0],
                                  emission = [0.0, 0.0, 0.0, 1.0],
                                  shininess = 0.1*128)

materials['brass'] = Material(ambient=[0.329412, 0.223529, 0.027451, 1.0],
                              diffuse=[0.780392, 0.568627, 0.113725, 1.0],
                              specular=[0.992157, 0.941176, 0.807843, 1.0],
                              emission = [0.0, 0.0, 0.0, 1.0],
                              shininess = 0.21794872*128)

materials['bronze'] = Material(ambient=[0.2125, 0.1275, 0.054, 1.0],
                               diffuse=[0.714, 0.4284, 0.18144, 1.0],
                               specular=[0.393548, 0.271906, 0.166721, 1.0],
                               emission = [0.0, 0.0, 0.0, 1.0],
                               shininess = 0.2*128)

materials['polished bronze'] = Material(
    ambient =[0.25, 0.148, 0.06475, 1.0],
    diffuse =[0.4, 0.2368, 0.1036, 1.0],
    specular =[0.774597, 0.458561, 0.200621, 1.0],
    shininess =76.8)

materials['chrome'] = Material(ambient=[0.25, 0.25, 0.25, 1.0],
                               diffuse=[0.4, 0.4, 0.4, 1.0],
                               specular=[0.774597, 0.774597, 0.774597, 1.0],
                               emission = [0.0, 0.0, 0.0, 1.0],
                               shininess = 0.6*128)

materials['copper'] = Material(ambient=[0.19125, 0.0735, 0.0225, 1.0],
                               diffuse=[0.7038, 0.27048, 0.0828, 1.0],
                               specular=[0.256777, 0.137622, 0.086014, 1.0],
                               emission = [0.0, 0.0, 0.0, 1.0],
                               shininess = 0.1*128)

materials['polished copper'] = Material(
    ambient =[ 0.2295, 0.08825, 0.0275, 1.0 ],
    diffuse =[0.5508, 0.2118, 0.066, 1.0 ],
    specular =[0.580594, 0.223257, 0.0695701, 1.0 ],
    shininess =51.2)
          
materials['gold'] =  Material( ambient=[0.24725, 0.1995, 0.0745, 1.0],
                               diffuse=[0.75164, 0.60648, 0.22648, 1.0],
                               specular=[0.628281, 0.555802, 0.366065, 1.0],
                               emission = [0.0, 0.0, 0.0, 1.0],
                               shininess = 0.4*128)

materials['polished gold'] = Material(
    ambient =[ 0.24725, 0.2245, 0.0645, 1.0 ],
    diffuse =[0.34615, 0.3143, 0.0903, 1.0 ],
    specular =[ 0.797357, 0.723991, 0.208006, 1.0],
    shininess =83.2)
          
materials['tin'] = Material(
    ambient =[ 0.105882, 0.058824, 0.113725, 1.0 ],
    diffuse =[0.427451, 0.470588, 0.541176, 1.0 ],
    specular =[0.333333, 0.333333, 0.521569, 1.0 ],
    shininess = 9.84615)

materials['silver'] = Material(
    ambient=[0.19225, 0.19225, 0.19225, 1.0],
    diffuse=[0.50754, 0.50754, 0.50754, 1.0],
    specular=[0.508273, 0.508273, 0.508273, 1.0],
    emission = [0.0, 0.0, 0.0, 1.0],
    shininess = 0.4*128)

materials['polished silver'] = Material(
    ambient =[ 0.23125, 0.23125, 0.23125, 1.0 ],
    diffuse =[0.2775, 0.2775, 0.2775, 1.0 ],
    specular =[0.773911, 0.773911, 0.773911, 1.0 ],
    shininess =89.6)

materials['black plastic'] = Material(ambient=[0.0, 0.0, 0.0, 1.0],
                                      diffuse=[0.01, 0.01, 0.01, 1.0],
                                      specular=[0.50, 0.50, 0.50, 1.0],
                                      shininess=.25*128)

materials['cyan plastic'] = Material(ambient=[0.0, 0.1, 0.06, 1.0],
                                     diffuse=[0.0, 0.50980392, 0.50980392, 1.0],
                                     specular=[0.50196078, 0.50196078,
                                               0.50196078, 1.0],
                                     shininess=.25*128)

materials['green plastic'] = Material(ambient=[0.0, 0.0, 0.0, 1.0],
                                      diffuse=[0.1, 0.35, 0.1, 1.0],
                                      specular=[0.45, 0.55, 0.45, 1.0],
                                      shininess=.25*128)

materials['red plastic'] = Material(ambient=[0.0, 0.0, 0.0, 1.0],
                                    diffuse=[0.5, 0.0, 0.0, 1.0],
                                    specular=[0.7, 0.6, 0.6, 1.0],
                                    emission=[0.0, 0.0, 0.0, 1.0],
                                    shininess=.25*128)

materials['white plastic'] = Material(ambient=[0.0, 0.0, 0.0, 1.0],
                                      diffuse=[0.55, 0.55, 0.55, 1.0],
                                      specular=[0.70, 0.70, 0.70, 1.0],
                                      shininess=.25*128)

materials['yellow plastic'] = Material(ambient=[0.0, 0.0, 0.0, 1.0],
                                       diffuse=[0.5, 0.5, 0.0, 1.0],
                                       specular=[0.60, 0.60, 0.50, 1.0],
                                       shininess=.25*128)

materials['black rubber'] = Material(ambient=[0.02, 0.02, 0.02, 1.0],
                                     diffuse=[0.01, 0.01, 0.01, 1.0],
                                     specular=[0.4, 0.4, 0.4, 1.0],
                                     emission=[0.0, 0.0, 0.0, 1.0],
                                     shininess=.078125*128)

materials['cyan rubber'] = Material(ambient=[0.0, 0.05, 0.05, 1.0],
                                    diffuse=[0.4, 0.5, 0.5, 1.0],
                                    specular=[0.04, 0.7, 0.7, 1.0],
                                    emission=[0.0, 0.0, 0.0, 1.0],
                                    shininess=.078125*128)

materials['green rubber'] = Material(ambient=[0.0, 0.05, 0.0, 1.0],
                                     diffuse=[0.4, 0.5, 0.4, 1.0],
                                     specular=[0.04, 0.7, 0.04, 1.0],
                                     shininess=.078125*128)

materials['red rubber'] = Material(ambient=[0.05, 0.0, 0.0, 1.0],
                                   diffuse=[0.5, 0.4, 0.4, 1.0],
                                   specular=[0.7, 0.04, 0.04, 1.0],
                                   shininess=.078125*128)

materials['white rubber'] = Material(ambient=[0.05, 0.05, 0.05, 1.0],
                                     diffuse=[0.5, 0.5, 0.5, 1.0],
                                     specular=[0.7, 0.7, 0.7, 1.0],
                                     emission=[0.0, 0.0, 0.0, 1.0],
                                     shininess=.078125*128)

materials['yellow rubber'] = Material(ambient=[0.05, 0.05, 0.0, 1.0],
                                      diffuse=[0.5, 0.5, 0.4, 1.0],
                                      specular=[0.7, 0.7, 0.04, 1.0],
                                      shininess=.078125*128)


## "throwaway" class for keeping track of object stuff
class GeomObject:
    pass

        
## class to provide openGL drawing functionality
class pygletDraw(drawable.Drawable):
    """
    yapCAD ``drawable`` subclass for OpenGL rendering with pyglet
    """

    def window(self):
        window = []
        try:
            # Try and create a window with multisampling (antialiasing)
            config = gl.Config(sample_buffers=1, samples=4, \
                               depth_size=16, double_buffer=True)
            window = pyglet.window.Window(resizable=True, config=config)
        except pyglet.window.NoSuchConfigException:
            # Fall back to no multisampling for old hardware
            window = pyglet.window.Window(resizable=True)
        self.__fps_display = pyglet.window.FPSDisplay(window)
        return window
    
        
    def glSetup(self):
        gl.glClearColor(0.2, 0.2, 0.3, 1)
        gl.glColor3f(1, 1, 1)
        gl.glEnable(gl.GL_DEPTH_TEST)
        gl.glEnable(gl.GL_CULL_FACE)
        gl.glEnable(gl.GL_NORMALIZE)
        
        gl.glLightfv(gl.GL_LIGHT0, gl.GL_POSITION, vec(50, 50, 10, 0))
        gl.glLightfv(gl.GL_LIGHT0, gl.GL_SPECULAR, vec(5, 5, 10, 1))
        gl.glLightfv(gl.GL_LIGHT0, gl.GL_DIFFUSE, vec(1, 1, 1, 1))
        gl.glLightfv(gl.GL_LIGHT1, gl.GL_AMBIENT, vec(0, 0, 0, 1.0))
        gl.glLightfv(gl.GL_LIGHT1, gl.GL_SPECULAR, vec(0.6, 0.6, 0.6, 1.0))
        gl.glLightfv(gl.GL_LIGHT1, gl.GL_DIFFUSE, vec(0.8, 0.8, 0.8, 1))
        gl.glLightfv(gl.GL_LIGHT1, gl.GL_POSITION, vec(-10.0, -20.0, 20.0, 0))

        # Create a default material and Group for the Model
        material = materials['default'].material
        self.group = pyglet.model.MaterialGroup(material=material)

        # Create a Material and Group for the ground plane
        material2 = materials['groundplane'].material
        self.group2 = pyglet.model.MaterialGroup(material=material2)

    ## utility funtion to update bounding box based on a list of points
    def upbb(self,pp,box,offset=[0,0,0,1]):
        epsP = point(epsilon,epsilon,epsilon)
        epsM = point(-epsilon,-epsilon,-epsilon)
        for p in pp:
            p = add(p,offset)
            if box:
                box = [ [min(box[0][0],p[0]),min(box[0][1],p[1]),min(box[0][2],p[2])],
                        [max(box[1][0],p[0]),max(box[1][1],p[1]),max(box[1][2],p[2])] ]
            else:
                box = [ add(p,epsM),add(p,epsP) ]
        return box
        

    def addSurface(self,s,batch,group,offset,bbx):
        assert s[0] == 'surface'
        vert = s[1]
        norm = s[2]
        ind = s[3]
        pp = []
        ln = int(len(vert)/3)
        for i in range(ln):
            pp.append(vert[i*3:(i+1)*3])
        bbx = self.upbb(pp,bbx,offset)
        batch.add_indexed(int(len(vert)/3),
                          gl.GL_TRIANGLES,
                          group,
                          ind,
                          ('v3f/static', vert),
                          ('n3f/static', norm))
        return bbx

    __immediate = {}
        
    def register_immediate(self,func):
        """ register a function for immediate-mode drawing"""
        self.__immediate[func.__name__] = func

    def makeBatches(self):
        # convert geometry lists to overall bounding box and
        # OpenGL-specific representations
        
        bbx = False
        # values for adding small offset to a point to make single-point bounding box
        if ( self.__points == [] and 
             self.__lines == [] and 
             self.__linestrips == [] and 
             len(self.__objectdict) == 0 and 
             self.__surfaces == [] ):
            raise ValueError('nothing to render')

        for p in self.__points:
            pp = p[0]
            pc = p[1]
            bbx = self.upbb(pp[1],bbx)
            self.__batch1.add(int(len(pp[1])/3),gl.GL_POINTS,self.group,pp,pc)

        for l in self.__lines:
            ll = l[0]
            lc = l[1]
            pp = [list(ll[1][0:3])]
            pp.append(list(ll[1][3:6]))
            bbx = self.upbb(pp,bbx)
            self.__batch1.add(int(len(ll[1])/3),gl.GL_LINES,self.group,ll,lc)
            
        for l in self.__linestrips:
            ll = l[0]
            lc = l[1]
            li = l[2]
            pp = []
            ln = int(len(ll[1])/3)
            for i in range(ln):
                pp.append(ll[1][i*3:(i+1)*3])
            bbx =self.upbb(pp,bbx)
            self.__batch1.add_indexed(int(len(ll[1])/3),
                                    gl.GL_LINES,
                                    self.group,
                                    li,ll,lc)

            

        def lineSurface(s,obj):
            assert s[0] == 'surface'
            vert = s[1]
            ind = s[3]
            drawn = []
            for j in range(0,len(ind),3):
                i = ind[j:(j+3)]
                l1 = i[0],i[1]
                if l1[0] > l1[1]:
                    l1 = i[1],i[0]
                l2 = i[1],i[2]
                if l2[0] > l2[1]:
                    l2 = i[2],i[1]
                l3 = i[2],i[0]
                if l3[0] > l3[1]:
                    l3 = i[0],i[2]
                if not l1 in drawn:
                    self.draw_line(vert[l1[0]*3:l1[0]*3+3],
                                   vert[l1[1]*3:l1[1]*3+3],
                                   entity=obj)
                    drawn.append(l1)
                if not l2 in drawn:
                    self.draw_line(vert[l2[0]*3:l2[0]*3+3],
                                   vert[l2[1]*3:l2[1]*3+3],
                                   entity=obj)
                    drawn.append(l2)
                if not l3 in drawn:
                    self.draw_line(vert[l3[0]*3:l3[0]*3+3],
                                   vert[l3[1]*3:l3[1]*3+3],
                                   entity=obj)
                    drawn.append(l3)
            for l in obj.lines:
                ll = l[0]
                lc = l[1]
                obj.linebatch.add(2,gl.GL_LINES,self.group,ll,lc)

                    
        for o in self.__objectdict.values():
            for s in o.surfaces:
                bbx = self.addSurface(s,o.batch,o.group,
                                      point(o.x,o.y,o.z),bbx)
                lineSurface(s,o)
            for l in o.lines:
                ll = l[0]
                lc = l[1]
                o.linebatch.add(2,gl.GL_LINES,self.group,ll,lc)

                

        for s in self.__surfaces:
            bbx = self.addSurface(s,self.__batch2,self.group,
                                  point(0,0,0),bbx)

        ## Create a ground plane
        self.__batch3.add_indexed(4,
                                gl.GL_TRIANGLES,
                                self.group2,
                                [0,1,2,2,3,0],
                                ('v3f/static',[bbx[0][0]*1.2,bbx[0][1]*1.2,bbx[0][2]-0.1,
                                               bbx[1][0]*1.2,bbx[0][1]*1.2,bbx[0][2]-0.1,
                                               bbx[1][0]*1.2,bbx[1][1]*1.2,bbx[0][2]-0.1,
                                               bbx[0][0]*1.2,bbx[1][1]*1.2,bbx[0][2]-0.1]),
                                ('n3f/static',[0,0,1]*4))
                              
                                  
                                
        self.__bbox = bbx
        rnge = sub(self.__bbox[1],self.__bbox[0])
        mdim = max(rnge)
        mdim = max(mdim,10)

        # print("scene bounding box: ",vstr(bbx))

        
    def __init__(self):

        super().__init__()
        # these layers added for compatibility with ezdxf_drawable
        # default layer list.  For now, selecting a layer has
        # no effect on drawing
        self.layerlist = [False, '0','PATHS','DRILLS','DOCUMENTATION']

        def on_key_press(symbol,modifiers):
            handled = False
            if symbol == pyglet.window.key.UP:
                handled = True
                self.__cameradist *= 1.1
                if self.__cameradist > self.__maxcameradist:
                    self.__cameradist = self.__maxcameradist
            elif symbol == pyglet.window.key.DOWN:
                handled = True
                self.__cameradist *= .90909090
                if self.__cameradist < self.__mincameradist:
                    self.__cameradist = self.__mincameradist
            elif symbol == pyglet.window.key.RETURN:
                handled = True
                self.__cameradist = self.camerastartdist
                self.__rx = 0
                self.__ry = 0
            elif symbol == pyglet.window.key.P:
                handled = True
                self.__drawground = not self.__drawground
            elif symbol == pyglet.window.key.M:
                handled = True
                self.__legend = not self.__legend
            elif symbol == pyglet.window.key.L:
                handled = True
                if not (self.__light0 or self.__light1):
                    self.__light0 = True
                    gl.glEnable(gl.GL_LIGHT0)
                    gl.glDisable(gl.GL_LIGHT1)
                elif self.__light0 and not self.__light1:
                    self.__light1 = True
                    gl.glEnable(gl.GL_LIGHT0)
                    gl.glEnable(gl.GL_LIGHT1)
                elif self.__light0 and self.__light1:
                    self.__light0 = self.__light1 = False
            if handled:
                return pyglet.event.EVENT_HANDLED
            else:
                return
            
                    
        self.__window = self.window()
        self.__window.push_handlers(on_key_press)
        self.__window.projection = pyglet.window.Projection3D(zfar=1000.0)
        self.glSetup()
        self.__center= point(0,0)
        self.__magnify = 0.08
        self.__arcres = 5
        self.__cameradist = self.camerastartdist = 100.0
        self.__maxcameradist = 900.0
        self.__mincameradist = 10.0
        self.__light0 = True
        self.__light1 = False
        self.__legend = True
        self.__drawground = True
        self.__rx = 0
        self.__ry = 0
        self.__points= []
        self.__lines= []
        self.__linestrips=[]
        self.__surfaces=[]
        self.__labels = []
        self.__batch1 = graphics.Batch() # use for lines, points, etc.
        self.__batch2 = graphics.Batch() # default, use for surfaces
        self.__batch3 = graphics.Batch() # use for environmental features, e.g. ground plane
        self.__objectdict = {} # use for animated objects
        self.__bbox = False

        @self.__window.event
        def on_mouse_drag(x, y, dx, dy, buttons, modifiers):
            self.__rx += dx
            self.__ry -= dy
            return pyglet.event.EVENT_HANDLED

        @self.__window.event
        def on_draw():
            self.__window.clear()
            gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
            
            self.__window.projection = pyglet.window.Projection3D(zfar=1000.0)
            # gl.glEnable(gl.GL_BLEND)
            gl.glColor3f(1., 1., 1.)

            #gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glLoadIdentity()
            if not self.__bbox:
                self.__bbox =[point(-1,-1,-1),point(1,1,1)]
            cent = scale3(add(self.__bbox[0],self.__bbox[1]),0.5)
            rnge = sub(self.__bbox[1],self.__bbox[0])
            mdim = max(rnge)
            mdim = max(mdim,10)
            
            gl.glTranslatef(-cent[0],-cent[1],-cent[2]-1*self.__cameradist)
            #gl.glTranslatef(0,0,-1*self.__cameradist)
            gl.glRotatef(self.__ry%360.0, 1, 0, 0)
            gl.glRotatef(self.__rx%360.0, 0, 1, 0)

            #draw "ground plane"
            if self.__drawground:
                gl.glEnable(gl.GL_LIGHTING)
                gl.glEnable(gl.GL_LIGHT0)
                self.__batch3.draw()  

            gl.glDisable(gl.GL_LIGHTING)
            self.__batch1.draw()
            
            if self.__light0 or self.__light1:
                gl.glEnable(gl.GL_LIGHTING)
                self.__batch2.draw()

            # execute any immediate-mode registered rendering functions
            for f in self.__immediate.values():
                f()
                
            for obj in self.__objectdict.values():
                gl.glMatrixMode(gl.GL_MODELVIEW)
                gl.glPushMatrix()
                gl.glTranslatef(obj.x,obj.y,obj.z)
                gl.glRotatef(obj.rx, 1, 0, 0)
                gl.glRotatef(obj.ry, 0, 1, 0)
                gl.glRotatef(obj.rz, 0, 0, 1)
                if obj.lighting and (self.__light0 or self.__light1):
                    gl.glEnable(gl.GL_LIGHTING)
                    obj.batch.draw()
                else:
                    gl.glDisable(gl.GL_LIGHTING)
                    obj.linebatch.draw()
                gl.glPopMatrix()

            #gl.glDisable(gl.GL_BLEND)
            gl.glDisable(gl.GL_LIGHTING)
            gl.glColor3f(1., 1., 1.)
            gl.glScalef(0.05, 0.05, 0.05)
            for label in self.__labels:
                gl.glPushMatrix()
                gl.glTranslatef(0,0,label[1])
                label[0].draw()
                gl.glPopMatrix()

            if self.__legend:
                gl.glMatrixMode(gl.GL_MODELVIEW)
                gl.glPushMatrix()
                gl.glLoadIdentity()
                gl.glMatrixMode(gl.GL_PROJECTION)
                gl.glPushMatrix()
                gl.glLoadIdentity()

                self.__window.projection = pyglet.window.Projection2D()
                label = pyglet.text.HTMLLabel(yapCAD_legend,
                                              x=10, y=30,
                                              anchor_x='left',anchor_y='bottom',
                                              width=400,
                                              multiline=True)
                label.draw()
                self.__fps_display.draw()
                gl.glMatrixMode(gl.GL_PROJECTION)
                gl.glPopMatrix()
                gl.glMatrixMode(gl.GL_MODELVIEW)
                gl.glPopMatrix()
         
            return pyglet.event.EVENT_HANDLED
            

    def __repr__(self):
        return 'an instance of pygletDraw'

    ## properties

    @property
    def magnify(self):
        return self.__magnify

    def _set_magnify(self,mag):
        self.__magnify=mag

    @magnify.setter
    def magnify(self,mag=False):
        if not isinstance(mag,(int,float)):
            raise ValueError('invalid magnification ' + str(mag))
        if isinstance(mag,bool) and mag ==False:
            mag = 1.0
        elif mag < epsilon:
            mag = epsilon
        self._set_magnify(mag)

    @property
    def cameradist(self):
        return self.__cameradist

    def _set_cameradist(self,dist):
        self.__cameradist=dist

    @cameradist.setter
    def cameradist(self,dist=False):
        if not isinstance(dist,(int,float)):
            raise ValueError('invalid camera distance ' + str(dist))
        if isinstance(dist,bool) and dist ==False:
            dist = 100.0
        elif dist < self.__mincameradist:
            dist = self.__mincameradist
        self._set_cameradist(dist)


    @property
    def rx(self):
        return self.__rx

    @rx.setter
    def rx(self,v):
        if not isinstance(v,(int,float)):
            raise ValueError('invalid camera rotation')
        self.__rx = v%360.0

    @property
    def ry(self):
        return self.__ry

    @ry.setter
    def ry(self,v):
        if not isinstance(v,(int,float)):
            raise ValueError('invalid camera rotation')
        self.__ry = v%360.0

    @property
    def objectdict(self):
        return self.__objectdict

    
    ## Overload virtual yapcad.drawable base class dawing methods
    
    def draw_point(self,p):
        ar = self.__arcres
        self.__arcres=30
        super().draw_point(p)
        self.__arcres=ar

    def draw_line(self,p1,p2,entity=None,c1=None,c2=None):

        if not entity:
            entity=self
            elist = entity.__lines
        else:
            elist = entity.lines
        color1 = []
        color2 = []
        if not c1:
            color1 = self.thing2color(entity.linecolor,'f')
        else:
            color1 = self.thing2color(c1,'f')
        if not c2:
            color2 = self.thing2color(entity.linecolor,'f')
        else:
            color2 = self.thing2color(c2,'f')
        elist.append([ ('v3f',(p1[0],p1[1],p1[2],
                               p2[0],p2[1],p2[2])),
                       ('c3f',tuple(color1 + color2)) ])
            
    def draw_arc(self,p,r,start,end):
        res = self.__arcres
        # resolution of arc sampling in degrees
        points = []
        if not (start==0 and end==360):
            start = start%360.0
            end = end % 360.0
            if end < start:
                end = end + 360
        theta = start*pi2/360.0
        points.append(add(p,[math.cos(theta)*r,math.sin(theta)*r,0.0,1.0]))
        for a in range(round(start),round(end),res):
            theta = a*pi2/360.0
            pp = [math.cos(theta)*r,math.sin(theta)*r,0.0,1.0]
            pp = add(pp,p)
            points.append(pp)
        theta = end*pi2/360.0
        points.append(add(p,[math.cos(theta)*r,math.sin(theta)*r,0.0,1.0]))

        self.draw_linestrip(points)

    def draw_text(self,text,location,
                  align='left',
                  attr={'name': 'Times New Roman',
                        'size': 12}):
        name = 'Times New Roman'
        if 'font_name' in attr:
            name = attr['font_name']
        bold = False
        if 'bold' in attr:
            bold = attr['bold']
        italic = False
        if 'italic' in attr:
            italic = attr['italic']
        underline = False
        if 'underline' in attr:
            underline = attr['underline']
        size = 12
        if 'size' in attr:
            size = attr['size']
        elif 'height' in attr:
            size *= attr['height']
        anchor_x='left'
        if 'anchor_x' in attr:
            anchor_x = attr['anchor_x']
        anchor_y='center'
        if 'anchor_y' in attr:
            anchor_y = attr['anchor_y']
        col = []
        if 'color' in attr:
            col = attr['color']
        elif self.linecolor:
            col = self.thing2color(self.linecolor,'b')
            # print ("color: ",col)
        else:
            col = [255,255,255]
            
        color = tuple(col + [255])
            
        x = location[0]
        y = location[1]
        
        self.__labels.append([pyglet.text.Label(text,
                                                font_name=name,
                                                font_size=size*self.__magnify,
                                                x=x*20.0, #fudge factor to make font rendering look OK
                                                y=y*20.0,
                                                align=align.lower(),
                                                color=color,
                                                bold=bold,
                                                italic=italic,
                                                #underline=underline,
                                                anchor_x=anchor_x,
                                                anchor_y=anchor_y),
                              location[2]*20.0])
        
    ## OpenGL-specific drawing methods
    
    def draw_linestrip(self,points):
        # we simulate a linestrip useing GL_LINES and indexed drawing.
        # This prevents extra lines
        p2 = []
        i2 = []
        color = self.thing2color(self.linecolor,'f')
        for p in points:
            p2 = p2 + p[0:3]
        for i in range(1,len(points)):
            i2 = i2 + [i-1, i]
        c2 = color * int(len(p2)/3)
        self.__linestrips.append([ ('v3f', tuple(p2)),\
                                   ('c3f',tuple(c2)),
                                   tuple(i2) ])

    def make_object(self,name,**kwargs):
        """
        Create a new three-dmensional object with specified
        material properties and 6DOF in world-space that can be
        animated
        """
        if name in self.__objectdict.keys():
            raise ValueError('duplicate object key')
        obj = GeomObject()

        if 'lighting' in kwargs:
            if kwargs['lighting']:
                obj.lighting = True
            else:
                obj.lighting=False
        if 'position' in kwargs:
            x = kwargs['position']
            if not ispoint(x):
                raise ValueError('bad position')
            obj.x = x[0]
            obj.y = x[1]
            obj.z = x[2]
        else:
            obj.x = 0
            obj.y = 0
            obj.z = 0
        if 'rotation' in kwargs:
            r = kwargs['rotation']
            obj.rx = r[0]
            obj.ry = r[1]
            obj.rz = r[2]
        else:
            obj.rx = 0
            obj.ry = 0
            obj.rz = 0
        if 'material' in kwargs:
            mat = kwargs['material']
            if mat in materials:
                obj.material = materials[mat]
                m = obj.material.material
                obj.group = pyglet.model.MaterialGroup(material=m)
            else:
                raise ValueError('unknown material')
        else:
            obj.material=materials['default']
            obj.group = self.group

        if 'animate' in kwargs:
            pyglet.clock.schedule_interval(kwargs['animate'],1/60)

        if 'linecolor' in kwargs:
            obj.linecolor = kwargs['linecolor']
        else:
            obj.linecolor = 'gray'

        obj.batch = graphics.Batch()
        obj.linebatch = graphics.Batch()
        obj.surfaces = []
        obj.lines=[]
        

        self.__objectdict[name] = obj
        
    def draw_surface(self,points,normals=None,faces=None,name=None):
        """
        render a surface, either as part of default scene geometry
        (if ``name`` is not specified) or as a part of a named object.
        """
        if not (normals or faces):
            if not issurface(points):
                raise ValueError('bad surface passed to draw_surface')
            normals = points[2]
            faces = points[3]
            points = points[1]
        vrts = []
        nrms= []
        inds = []
        for p in points:
            vrts+=p[0:3]
        for n in normals:
            nrms+=n[0:3]
        for f in faces:
            inds+=f[0:3]
        if not name:
            self.__surfaces.append(['surface',vrts,nrms,inds])
        else:
            obj = self.__objectdict[name]
            obj.surfaces.append(['surface',vrts,nrms,inds])

    def draw_solid(self,solid,name=None):
        """
        Render a solid, either as part of the secene geometry
        """
        if not issolid(solid):
            raise ValueError('bad solid passed to draw_solid')
        for surface in solid[1]:
            self.draw_surface(surface,name=name)

    ## override base-class draw method
    def draw(self,x,name=None):
        if issolid(x):
            self.draw_solid(x,name=name)
        elif issurface(x):
            self.draw_surface(x,name=name)
        else:
            super().draw(x)

    ## override base-class virtual display method
    def display(self):
        self.makeBatches()
        pyglet.app.run()




## boxcut procedural design example for yapCAD
## Born on 12 December, 2020
## Copyright (c) 2020 Richard DeVaul

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""procedural design example that generates squeeze-fit box designs from
input parameters and produces DXF files as output.

This is intended as a demonstration design system -- more work would
be needed to generalize this to designs bigger than 200mm or so on a
side.

"""

import math
from yapcad.geom import *
from yapcad.geom3d import *
from yapcad.geom3d_util import *
from yapcad.poly import *
from yapcad.combine import *
from yapcad.ezdxf_drawable import *

## global configuration variables

# minimum box dimension, mm
# NOTE: this is dictated by the tab structure
mindim = 65.0

#tab length, mm
tlen = 6.0

## global state variables, set in __main__

# box dimensions, nominal default values.  Internally, all dimensions
# are in mm.
box_length = mindim
box_width = mindim
box_height = mindim

# base filename for DXF generation
outfile = "boxout"

# material thickness, mm
thick = 3.175

# kerf compensation, mm
kerf = 0.397

# do we do openGL?
ogl = False

# a global geometry list we use for rendering debug-useful geometry
debugGeomList=[]

renderDebug=False

# define the bottom and top pieces by outline.  These share the same
# edge-join geometry:

#      +----__--__------__--__----+ 
#      |                          | 
#      |                      ^   | 
#       |  Bottom and Top     |  |  
#      |                      |   | 
#       |  Tab pattern is     |  |    
#      |   the same on      width | 
#      |   all edges.         |   |
#       |                     |  |  
#      | <------ length ------+-> |
#       |                     |  |
#      |                      v   |
#      |    __  __      __  __    |
#      +----  --  ------  --  ----+

def topPoly(toplen=box_length,topwid=box_width):
    # start with a rounded rectangle the size of the face
    face = RoundRect(toplen+kerf*2,topwid+kerf*2,2.0)

    debugGeomList.append(face.geom)
    # make trimming polygons
    trimtab = Rect(tlen-kerf*2,(thick-kerf)*2)
    trimtab2 = Rect((thick-kerf)*2,tlen-kerf*2)

    tl2 = toplen/2.0
    tw2 = topwid/2.0
    t1off = tlen*2.5
    t2off = tlen*4.5

    # duplicate and position to trim the upper, right corner
    t1a = deepcopy(trimtab)
    t1a.translate(point(tl2-t1off, tw2))

    t1b = deepcopy(trimtab)
    t1b.translate(point(tl2-t2off, tw2))

    t2a = deepcopy(trimtab2)
    t2a.translate(point(tl2, tw2-t1off))

    t2b = deepcopy(trimtab2)
    t2b.translate(point(tl2, tw2-t2off))

    ## Add these features to debugging geometry list
    debugGeomList.append(t1a.geom)
    debugGeomList.append(t1b.geom)
    debugGeomList.append(t2a.geom)
    debugGeomList.append(t2b.geom)
    
    # perform boolean operations for each corner -- use mirror to
    # rotate the trimming bodies around the rectangular form.

    for i in range(4):
        if i == 0:
            pass # we are in correct position for first trim
        elif i == 1 or i == 3:
            t1a.mirror("yz")
            t1b.mirror("yz")
            t2a.mirror("yz")
            t2b.mirror("yz")
        else: # i == 2
            t1a.mirror("xz")
            t1b.mirror("xz")
            t2a.mirror("xz")
            t2b.mirror("xz")

        face = Boolean('difference',[face,t1a])
        face = Boolean('difference',[face,t1b])
        face = Boolean('difference',[face,t2a])
        face = Boolean('difference',[face,t2b])

    return face

# define the front and back pieces by outline.  These share the same
# edge-join geometry:

#           __  __      __  __      
#      +----  --  ------  -- ^----+ 
#      |                     |    | 
#       |  Front and Back    |   |  
#      |                     |    | 
#       |  Tab pattern is    |   |    
#      |   the same on     height | 
#      |   opposite edges    |    |
#       |                    |   |  
#      | <------ length -----+--> |
#       |                    |   |
#      |                     v    |
#      +----__--__------__--__----+
#                                  

def frontPoly(lrlen=box_length,lrht=box_height):
    # start with a rounded rectangle the size of the face, less the
    # projecting tabs
    face = RoundRect(lrlen+kerf*2,lrht+(kerf-thick)*2,2.0)

    #debugGeomList.append(face.geom)
    # make tab and trim polygons
    tab = Rect(tlen+kerf*2,(thick+kerf)*2)
    trimtab = Rect((thick+kerf)*2,tlen+kerf*2)

    tl2 = lrlen/2.0
    tw2 = lrht/2.0
    t1off = tlen*2.5
    t2off = tlen*4.5

    # duplicate and position to trim the upper, right corner
    t1a = deepcopy(tab)
    t1a.translate(point(tl2-t1off, tw2-thick))
    
    t1b = deepcopy(tab)
    t1b.translate(point(tl2-t2off, tw2-thick))
    
    t2a = deepcopy(trimtab)
    t2a.translate(point(tl2, tw2-t1off))
    
    t2b = deepcopy(trimtab)
    t2b.translate(point(tl2, tw2-t2off))

    #debugGeomList.append(t1a.geom)
    #debugGeomList.append(t1b.geom)
    #debugGeomList.append(t2a.geom)
    #debugGeomList.append(t2b.geom)
    
    # perform boolean operations for each corner -- use mirror to
    # rotate the trimming bodies around the rectangular form.

    for i in range(4):
        if i == 0:
            pass # we are in correct position for first trim
        elif i == 1 or i == 3:
            t1a.mirror("yz")
            t1b.mirror("yz")
            t2a.mirror("yz")
            t2b.mirror("yz")
        else: # i == 2
            t1a.mirror("xz")
            t1b.mirror("xz")
            t2a.mirror("xz")
            t2b.mirror("xz")
            
        face = Boolean('union',[face,t1a])
        face = Boolean('union',[face,t1b])
        face = Boolean('difference',[face,t2a])
        face = Boolean('difference',[face,t2b])

    return face

# define the left and right pieces by outline.  These share the same
# edge-join geometry:

#           __  __      __  __      
#       +---  --  ------  -- ^---+  
#       |                    |   |  
#      |   Left and Right    |    | 
#       |                    |   |  
#      |   Tab pattern is    |    |   
#       |  the same on     height|  
#       |  opposite edges    |   | 
#      |                     |    | 
#       |                    |   | 
#      |<-------- width -----+--->|
#       |                    v   | 
#       +---__--__------__--__---+ 
#                                  

def leftPoly(lrwid=box_width,lrht=box_height):
    # start with a rounded rectangle the size of the face, less the
    # projecting tabs
    face = RoundRect(lrwid+(kerf-thick)*2,lrht+(kerf-thick)*2,2.0)

    #debugGeomList.append(face.geom)
    # make tab and trim polygons
    tab1 = Rect(tlen+kerf*2,(thick+kerf)*2)
    tab2 = Rect((thick+kerf)*2,tlen+kerf*2)

    tl2 = lrwid/2.0
    tw2 = lrht/2.0
    t1off = tlen*2.5
    t2off = tlen*4.5

    # duplicate and position to add tabs to the upper, right corner
    t1a = deepcopy(tab1)
    t1b = deepcopy(tab1)
    t1a.translate(point(tl2-t1off, tw2-thick))
    t1b.translate(point(tl2-t2off, tw2-thick))
    t2a = deepcopy(tab2)
    t2b = deepcopy(tab2)
    t2a.translate(point(tl2-thick, tw2-t1off))
    t2b.translate(point(tl2-thick, tw2-t2off))

    # add these to our debugging geometry list for later
    # visualization, if desired
    #debugGeomList.append(t1a.geom)
    #debugGeomList.append(t1b.geom)
    #debugGeomList.append(t2a.geom)
    #debugGeomList.append(t2b.geom)
    
    # perform boolean operations for each corner -- use mirror to
    # rotate the trimming bodies around the rectangular form.

    for i in range(4):
        if i == 0:
            pass # we are in correct position for first corner
        elif i == 1 or i == 3:
            t1a.mirror("yz")
            t1b.mirror("yz")
            t2a.mirror("yz")
            t2b.mirror("yz")
        else: # i == 2
            t1a.mirror("xz")
            t1b.mirror("xz")
            t2a.mirror("xz")
            t2b.mirror("xz")
            
        face = Boolean('union',[face,t1a])
        face = Boolean('union',[face,t1b])
        face = Boolean('union',[face,t2a])
        face = Boolean('union',[face,t2b])

    face.rotate(90)
    return face

def makePolys():
    polylist = []

    top = topPoly(box_length,box_width)
    front = frontPoly(box_length,box_height)
    left = leftPoly(box_width,box_height)
    
    polylist.append(top)
    polylist.append(front)
    polylist.append(left)

    return polylist

## Draw some documentary text using the specified drawable.
## Text in the OpenGL rendering isn't working very well yet.

def legend(d):

    x = (box_length/2)*1.2
    y = (box_width/2+box_height)
    
    d.draw_text("yapCAD", point(x,y),\
                attr={'style': 'OpenSans-Bold', # style for ezdxf
                      'font_name': 'OpenSans', # style for pyglet
                      'bold': True, # style for pyglet
                      'height': 20.0})
    y -= 25
    d.draw_text("boxcut.py",
                point(x,y),
                attr={'height': 15.0})

    y -= 20
    d.draw_text("length: {:.2f}mm".format(box_length),
                point(x,y),
                attr={'height': 6.0})
    y -= 10
    d.draw_text("width: {:.2f}mm".format(box_width),
                point(x,y),
                attr={'height': 6.0})
    y -= 10
    d.draw_text("height: {:.2f}mm".format(box_height),
                point(x,y),
                attr={'height': 6.0})

    y -= 10
    d.draw_text("kerf: {:.3f}mm".format(kerf),
                point(x,y),
                attr={'height': 6.0})


def twoDDraw(polys,dd,zoff=0.0):

    dd.layer = 'DOCUMENTATION'
    dd.linecolor = 'yellow'
    legend(dd)

    # calculate some spacing intervals
    lenhei = (box_length+box_height)/2.0
    widhei = (box_width+box_height)/2.0
    
    for i in range(len(polys)):
        p = polys[i]
        d1 = []
        d2 = []
        d1s = ""
        d2s = ""
        if i == 0:
            d1 = point(0,0,zoff) # location of bottom
            d2 = point(lenhei*1.1*2,0,zoff) # location of top
            d1s = "bottom"
            d2s = "top"
        elif i == 1: 
            d1 = point(0,-widhei*1.1,zoff) # location of front
            d2 = point(0,widhei*1.1,zoff) # location of back
            d1s = "front"
            d2s = "back"
        else: # i == 2
            d1 = point(-lenhei*1.1,0,zoff) #location of left
            d2 = point(lenhei*1.1,0,zoff) #location of right
            d1s = "left"
            d2s = "right"
            
        dd.layer = 'PATHS'
        dd.linecolor = 'white'

        g = p.geom
        g1 = translate(g,d1)
        g2 = translate(g,d2)
        
        dd.draw(g1)
        dd.draw(g2)

        dd.layer = 'DOCUMENTATION'
        dd.linecolor = 'yellow'
        
        # att = {'style': 'LiberationMono','height': 10.0}
        att = {'height': 6.0, 'anchor_x': 'center'}
        
        dd.draw_text(d1s,d1, align='CENTER', attr=att)
        dd.draw_text(d2s,d2, align='CENTER', attr=att)
    
    
    if renderDebug:
        dd.layer = 'DOCUMENTATION'
        dd.linecolor='aqua'
        dd.draw(debugGeomList)

## So, yapCAD currently doesn't support rotating arcs out of the XY
## plane. That sucks for our visualization.  To hack around this, I'm
## creating a couple of utility functions that will replace arcs in
## geometry lists with sampled polyline representations.

# utility function to convert an arc to a polyline
def arc2poly(c, res=10.0):
    # resolution of arc sampling in degrees
    points = []
    p = c[0]
    r = c[1][0]
    start = c[1][1]
    end = c[1][2]
    if not (start==0 and end==360):
        start = start%360.0
        end = end % 360.0
        if end < start:
            end = end + 360
        theta = start*pi2/360.0
        points.append(add(p,[math.cos(theta)*r,math.sin(theta)*r,0.0,1.0]))
        for a in range(round(start),round(end),round(res)):
            theta = a*pi2/360.0
            pp = [math.cos(theta)*r,math.sin(theta)*r,0.0,1.0]
            pp = add(pp,p)
            points.append(pp)
        theta = end*pi2/360.0
        points.append(add(p,[math.cos(theta)*r,math.sin(theta)*r,0.0,1.0]))
    return points

# utility function to replace arcs in a geometry list with sampled
# polyline represntations
def glistArc2poly(glist):
    ngl = []
    for e in glist:
        if isarc(e):
            ngl.append(arc2poly(e))
        else:
            ngl.append(e)
    return ngl

# take an XY poly, extrude in z, return geometry list. NOTE: this does
# not produce a surface representation, only a wireframe of lines as a
# geometry list

def extrudeZ(poly,thick):
    geom = poly.geom
    geom = glistArc2poly(geom)
    # bottom = reverseGeomList(geom)
    bottom = geom
    bottom = translate(bottom,point(0,0,-thick/2))
    top = geom
    top = translate(top,point(0,0,thick/2))

    def glist2struts(glist):
        struts=[]
        upr = point(0,0,thick)
        for e in glist:
            strut=[]
            if ispoint(e):
                strut = line(e,add(e,upr))
                struts.append(strut)
            elif isline(e):
                strut = line(e[0],add(e[0],upr))
                struts.append(strut)
            elif isarc(e):
                p = samplearc(e,0.0)
                strut = line(p,add(p,upr))
                struts.append(strut)
            elif ispoly(e) or isgeomlist(e):
                struts = struts + glist2struts(e)
            else:
                raise ValueError('bad thing passed to glist2struts')
        return struts
    
    strts = glist2struts(bottom)
    return bottom + top + strts
    
    
def makeOglGeom():
    ## regenerate geometry with zero kerf correction
    global kerf
    oldkerf = kerf
    kerf = 0.0
    polys = makePolys()
    kerf = oldkerf

    parts = []
    wrfrm = []
    for poly in polys:
        # import pdb ; pdb.set_trace()
        ply = geomlist2poly(poly.geom)
        ply = translate(ply,point(0,0,-thick/2))
        surf,bnd = poly2surfaceXY(ply)
        part = extrude(surf,thick)
        part2 = extrudeZ(poly,thick)
        parts.append(part)
        wrfrm.append(part2)

    model = []

    def mkmdl(prts):
        modl=[]
        for i in range(len(prts)):
            part = prts[i]
            comp1 = []
            comp2 = []
            if i == 0: # top and bottom
                comp1 = translate(part,point(0,0,-(box_height-thick/2)/2))
                comp2 = translate(part,point(0,0,(box_height-thick/2)/2))
            elif i == 1: # front and back
                part = rotate(part,90.0,axis=point(1,0,0))
                comp1 = translate(part,point(0,-(box_width-thick/2)/2,0))
                comp2 = translate(part,point(0,(box_width-thick/2)/2,0))
            else: # i = 2, left and right
                part = rotate(part,90.0,axis=point(0,1,0))
                comp1 = translate(part,point(-(box_length-thick/2)/2,0,0))
                comp2 = translate(part,point((box_length-thick/2)/2,0,0))
            modl.append(comp1)
            modl.append(comp2)
        return modl

    model = mkmdl(parts)
    model.append(mkmdl(wrfrm))
    
    return model
        
if __name__ == "__main__":
    import argparse
    
    parser = argparse.ArgumentParser(description="""
    procedural design example that generates squeeze-fit box designs from
    input parameters and produces DXF files as output.
    """)

    parser.add_argument("length",type=float, help="length of the box")
    parser.add_argument("width",type=float, help="width of the box")
    parser.add_argument("height",type=float, help="height of the box")
    parser.add_argument("-u","--units",type=str,choices=["mm","inch"],
                        default="mm",help="units for dimensions")
    parser.add_argument("-o","--outname",type=str,default="boxout",
                        help="file name base for output, default is boxout")
    parser.add_argument("-t","--thick",type=float,default=-1,
                        help="material thickness, default is 3.175 mm (1/4\")")
    parser.add_argument("-k","--kerf",type=float,default=-1,
                        help="kerf compensation, default is 0.397mm (1/64\")")
    parser.add_argument("-g","--gl",action="store_true",
                        help="do an interactive 3D visualization of the box")
    parser.add_argument("-d","--debug",action="store_true",
                        help="turn on rendering of debug geometry")
    args = parser.parse_args()
    print(args)

    dimmul = 1.0
    if args.units == "inch":
        dimmul = 25.4

    box_width = args.width*dimmul
    box_length = args.length*dimmul
    box_height = args.height*dimmul

    if box_width < mindim or box_length < mindim or box_height < mindim:
        raise ValueError('minimum length of {}mm required for each dimension'.format(mindim))

    outname = args.outname
    if outname == "":
        raise ValueError('non-zero length file name base required')
    
    if args.debug:
        renderDebug=True

    if args.thick > 0:
        thick = args.thick

    if args.kerf >= 0:
        kerf = args.kerf

    if args.gl:
        ogl = True

    polys = makePolys()
    dd = ezdxfDraw()
    dd.filename = outfile

    twoDDraw(polys,dd)
    dd.display()

    if ogl:
        try:
            from yapcad.pyglet_drawable import *
            dd2 = pygletDraw()
        except RuntimeError as err:
            print("\nSkipping 3D visualization: {}".format(err))
            ogl = False
        except Exception as err:
            print("\nSkipping 3D visualization due to unexpected error: {}".format(err))
            ogl = False

    if ogl:
        # magnification factor for text
        dd2.magnify=1.5
        # compute the camera distance assuming a 60 degree (pi/6) FOV
        maxd = max((box_length*2+box_height*2)*1.1,
                   (box_width+ box_height*2)*1.1)
        dist = ((maxd/2) / math.sin(pi/6))*math.cos(pi/6)
        dd2.cameradist = dist-box_height*2

        twoDDraw(polys,dd2,-box_height*2)

        oglGeom = makeOglGeom()

        dd2.layer='PATHS'
        for i in range(len(oglGeom)):
            if i == 0 or i == 1:
                dd2.linecolor = 'silver'
            elif i == 2 or i == 3:
                dd2.linecolor = 'aqua'
            else:
                dd2.linecolor = 'fuchsia'
            dd2.draw(oglGeom[i])
        dd2.display()
    print("done")



import pytest
import os
import random
from yapcad.geom import *
from yapcad.geom_util import *
from yapcad.geometry import *

"""test functions for the yapcad.geometry module"""

# Control flag for visual tests - set via environment variable or directly
VISUALTEST = os.environ.get('VISUALTEST', 'false').lower() in ('true', '1', 'yes')
if VISUALTEST:
    from yapcad.pyglet_drawable import pygletDraw

class TestGeometry:
    """Tests for the Geometry class"""
    ## make some test geometry

    # bounding box for tests
    bb1 = [point(-10,-10,0),
           point(10,10,0)]

    def test_geom(self):
        """test basic yapcad.geom figure wrapping"""

        a = point(-5,-1)
        b = point(5,3)
        L = Line(a,b)
        assert L.sampleable
        assert L.intersectable
        assert L.continuous
        assert not L.closed
        assert close(L.length,
                     sqrt(10*10+4*4))
        print("line sample tests")
        p0 =L.sample(-0.5)
        p1 =L.sample(0.0)
        p2 =L.sample(0.5)
        p3 =L.sample(1.0)
        p4 =L.sample(1.5)
        assert vclose(p0,point(-10,-3))
        assert vclose(p1,a)
        assert vclose(p2,point(0,1))
        assert vclose(p3,b)
        assert vclose(p4,point(10,5))
        print("line unsample tests")
        u0 = L.unsample(p0)
        u1 = L.unsample(p1)
        u2 = L.unsample(p2)
        u3 = L.unsample(p3)
        u4 = L.unsample(p4)
        assert close(u0,-0.5)
        assert close(u1,0.0)
        assert close(u2,0.5)
        assert close(u3,1.0)
        assert close(u4,1.5)
        p = point(100,100)
        assert not L.unsample(p)
        
    @pytest.mark.visual
    def test_visual_intersect(self):
        if not VISUALTEST:
            pytest.skip("Visual tests disabled (set VISUALTEST=true to enable)")

        dd = pygletDraw()
        dd.linecolor='silver'
        # make a polyline spiral
        Ply = Figure(makeLineSpiral(point(-5,0),2.0,10))
        # make an arc-segment geometry list spiral
        Gl = Figure(makeArcSpiral(point(5,0),3.0,6.66,dstep=90.0))
        # find intersection points
        pts = Gl.intersectXY(Ply)
        # find intersection parameters
        uu = Ply.intersectXY(Gl,params=True)

        dd.draw(Ply)
        dd.draw(Gl)
        dd.polystyle='points'
        dd.pointstyle='x'
        dd.linecolor='aqua'
        dd.draw(pts)
        dd.linecolor='white'
        pts2 = []
        for i in range(len(uu[0])):
            p0 =Gl.sample(uu[1][i])
            assert vclose(p0,pts[i])
            pts2.append(p0)
        dd.pointstyle='o'
        dd.draw(pts2)
        dd.display()

    @pytest.mark.visual
    def test_properties(self):
        if not VISUALTEST:
            pytest.skip("Visual tests disabled (set VISUALTEST=true to enable)")

        # make some randm polys
        box = line(point(-10,-10),
                   point(10,10))
        plys=[]
        nply = 3
        for i in range(nply):
            ply = randomPoly(box,maxr=3.0)
            Ply = Geometry(ply)
            plys.append(Ply)
            assert Ply.issampleable()
            assert Ply.isintersectable()
            assert Ply.iscontinuous()
            assert Ply.isclosed()

        npts = 3
        for i in range(npts):
            ply = randomPoints(box,i+2)
            Ply = Geometry(ply)
            plys.append(Ply)
            assert Ply.issampleable()
            assert Ply.isintersectable()
            assert Ply.iscontinuous()
            assert not Ply.isclosed()

        sprls= 3
        for i in range(sprls):
            r = 3.0
            x = randomCenterInBox(box,r)
            sp = makeArcSpiral(x,r/(i+1.0),i+1.0)
            if not i%2:
                sp.append(line(sample(sp,1.0),
                               sample(sp,0.0)))
            Ply = Geometry(sp)
            plys.append(Ply)
            assert Ply.issampleable()
            assert Ply.isintersectable()
            assert iscontinuousgeomlist(sp)
            assert Ply.iscontinuous()
            if not i%2:
                assert Ply.isclosed()
            else:
                assert not Ply.isclosed()

        dd = pygletDraw()
        dd.linecolor='silver'
        dd.polystyle='both'
        dd.draw(plys)
        dd.display()

    @pytest.mark.visual
    def test_surface(self):
        if not VISUALTEST:
            pytest.skip("Visual tests disabled (set VISUALTEST=true to enable)")

        dd = pygletDraw()
        dd.linecolor='silver'
        # make an arc-segment geometry list spiral
        gl = makeArcSpiral(point(-10,0),10.0,1.0)
        # "close" the arc with a line segment
        gl.append(line(sample(gl,1.0),
                       sample(gl,0.0)))
        Nautalus = Geometry(gl)
        dd.draw(Nautalus)
        # turn the geometry list into a poly
        surf = Nautalus.surface()
        dd.draw_surface(surf)
        Nautalus.translate(point(0,0,0.5))
        surf2 = Nautalus.surface()
        dd.linecolor='red'
        dd.draw(surf2lines(surf2))
        

        ply2 = [ point(20,20),
                 point(10,20),
                 point(10,-20),
                 point(20,-20),
                 point(20,-10),
                 point(15,-10),
                 point(15,10),
                 point(20,10),
                 point(20,20) ]

        BigC=Geometry(ply2)
        dd.linecolor='silver'
        dd.draw(BigC)
        surf = BigC.surface()
        dd.draw_surface(surf)
        BigC.translate(point(0,0,0.5))
        surf2 = BigC.surface()
        dd.linecolor='red'
        dd.draw(surf2lines(surf2))
        print(f"surface area: {surfacearea(surf)}")
        assert abs(surfacearea(surf)-300.0) < 0.1
        
        # ply2 = translate(ply2,point(0,0,0.5)) # pull it forward
        # dd.linecolor='red'
        # dd.polystyle='both'
        # dd.draw(ply2)
        # surf2,bnd = poly2surface(ply2)
        # dd.draw_surface(surf2)

        # rr1 = makeRoundRect(5,5,0.5,center=point(-15,20))
        # rr2 = makeRoundRect(6,3,0.5,center=point(-12,20))
        # gl3 = Boolean('difference',[rr1,rr2])
        # ply3 = geomlist2poly(gl3.geom)
        # dd.linecolor='aqua'
        # dd.draw(gl3)
        # ply3 = translate(ply3,point(0,0,0.5)) # pull it forward
        # surf3,bnd = poly2surface(ply3)
        # dd.linecolor='red'
        # dd.polystyle='both'
        # dd.draw(ply3)
        # dd.draw_surface(surf3)

        # gl4 = [arc(point(0,7),10.0,0.0,270.0),
        #        arc(point(0,7),5.0,0.0,270.0,samplereverse=True)]
        # gl4 = [arc(point(0,7),10.0,0.0,350.0),
        #        arc(point(0,7),5.0,0.0,350.0,samplereverse=True)]
        # dd.linecolor='aqua'
        # dd.draw(gl4)
        # ply4 =geomlist2poly(gl4,minang = 10 )
        # ply4.append(ply4[0])

        # ply4 = translate(ply4,point(0,0,0.5)) # pull it forward
        # dd.linecolor='red'
        # dd.polystyle='both'
        # dd.draw(ply4)

        # surf4,bnd = poly2surface(ply4,minlen=1.0)
        # # assert False
        # lines = surf2lines(surf4)
        # lines = translate(lines,point(0,0,0.5))
        # dd.draw_surface(surf4)
        # dd.linecolor='yellow'
        # dd.draw(lines)
        # bnd.append(bnd[0])
        # bnd = translate(bnd,point(0,0,1.0))
        # dd.linecolor='white'
        # dd.draw(bnd)
        
        dd.display()
        
    # def test_bool(self):
        
    #     dd = pygletDraw()
    #     dd.linecolor='silver'
    #     dd
        
                        



## quadtree and octtree representations for yapCAD geometry
## Born on 15 December 2020
## Copyright (c) 2020 Richard DeVaul

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""quadtree and octtree representations for yapCAD geometry"""

from yapcad.geom import *

# determine if two bounding boxes overlap.  This is a more challenging
# problem than it may appear at first glance.  It is not enough to
# test the corner points of one box to see if they fall inside the
# other, as illustrated by the following cases:

# case 1:  +--------------+
# inside   |     box1     |
#          |  +--------+  |
#          |  |  box2  |  |
#          |  +--------+  |
#          +--------------+

# No corner points of box1 lie inside box2, no lines intersect.

# case 2:     +------+
# cross       | box1 |
#     +-------+------+------+
#     | box2  |      |      |
#     +-------+------+------+
#             |      |
#             +------+

# No corner point of either box lie inside the other, projected lines
# intersect.

def boxoverlap2(bbx1,bbx2,dim3=True):
    """Determine if two bounding boxes overlap.  if dim3==True, treat the
    bounding boxes as 3D, otherwise treat them as co-planar 2D boxes.

    First, check to see if the maximum coordinates of one box are
    smaller than the minimum coordinates of the other, or vice versa.
    If so, no overlap is possible; return False

    if overlap is possible by test #1, check for the box-in-box
    special case for each box.  If so, return True

    Finally, for the 2D case: determine if horizontal lines of box1
    intersect with vertical lines of box2, and vice versa.  If any
    intersections found, return True, else return False

    For the 3D case, project the boxes into the XY, YZ, and XZ planes,
    and perform the 2D lines intersection check, as above.  Return
    True if and only if intersections are reported for each
    projection, otherwise return False
    """

    # check minmax
    if ((bbx1[1][0] < bbx2[0][0] and
         bbx1[1][1] < bbx2[0][1] and
         (not dim3 or bbx1[1][2] < bbx2[0][2])) or
        (bbx2[1][0] < bbx1[0][0] and
         bbx2[1][1] < bbx1[0][1] and
         (not dim3 or bbx2[1][2] < bbx1[0][2]))):
        return False # no overlap possible

    # check for box-in-box
    if ((bbx1[0][0] >= bbx2[0][0] and bbx1[1][0] <= bbx2[1][0] and
         bbx1[0][1] >= bbx2[0][1] and bbx1[1][1] <= bbx2[1][1] and
         (not dim3 or (bbx1[0][2] >= bbx2[0][2]
                       and bbx1[1][2] <= bbx2[1][2]))) or
        (bbx2[0][0] >= bbx1[0][0] and bbx2[1][0] <= bbx1[1][0] and
         bbx2[0][1] >= bbx1[0][1] and bbx2[1][1] <= bbx1[1][1] and
         (not dim3 or (bbx2[0][2] >= bbx1[0][2]
                       and bbx2[1][2] <= bbx1[1][2])))):
        return True

    def int2D(bb1,bb2,plane='XY'):
        """utility function for 2D box line intersection finding"""
        i = 0
        j = 0
        if plane == 'XY':
            i = 0
            j = 1
        elif plane == 'YZ':
            i = 1
            j = 2
        elif plane == 'XZ':
            i = 0
            j = 2
        else:
            raise ValueError('bad plane in int2D')

        def intHV(hln,vln):
            """ do a horizontal and vertial line intersect? """
            minx=hln[0][0]
            maxx=hln[1][0]
            if minx>maxx:
                swap = minx ; minx = maxx; maxx = swap

            if vln[0][0] < minx or vln[0][0] > maxx:
                return False
            miny=vln[0][1]
            maxy=vln[1][1]
            if miny > maxy:
                swap = miny ; miny = maxy ; maxy = swap
            if miny > hln[0][1] or maxy < hln[0][1]:
                return False
            return True
        
        # check for projected box-in-box
        if ((bb1[0][i] >= bb2[0][i] and bb1[1][i] <= bb2[1][i] and
             bb1[0][j] >= bb2[0][j] and bb1[1][j] <= bb2[1][j])
            or
            (bb2[0][i] >= bb1[0][i] and bb2[1][i] <= bb1[1][i] and
             bb2[0][j] >= bb1[0][j] and bb2[1][j] <= bb1[1][j])):
            return True

        # check for projected line intersections
        
        len1 = bb1[1][i] - bb1[0][i] # length
        wid1 = bb1[1][j] - bb1[0][j] # width
        len2 = bb2[1][i] - bb2[0][i] # length
        wid2 = bb2[1][j] - bb2[0][j] # width

        p0 = point(bb1[0][i],bb1[0][j])
        p1 = add(p0,point(len1,0))
        p2 = add(p1,point(0,wid1))
        p3 = add(p0,point(0,wid1))

        p4 = point(bb2[0][i],bb2[0][j])
        p5 = add(p4,point(len2,0))
        p6 = add(p5,point(0,wid2))
        p7 = add(p4,point(0,wid2))

        box1 = [[p0,p1],
                [p1,p2],
                [p2,p3],
                [p3,p0]]
        box2 = [[p4,p5],
                [p5,p6],
                [p6,p7],
                [p7,p4]]
        if intHV(box1[0],box2[1]):
            return True
        if intHV(box1[0],box2[3]):
            return True
        if intHV(box1[2],box2[1]):
            return True
        if intHV(box1[2],box2[3]):
            return True

        if intHV(box2[0],box1[1]):
            return True
        if intHV(box2[0],box1[3]):
            return True
        if intHV(box2[2],box1[1]):
            return True
        if intHV(box2[2],box1[3]):
            return True
        # if lineLineIntersectXY(box1[0],box2[1]):
        #     return True
        # if lineLineIntersectXY(box1[0],box2[3]):
        #     return True
        # if lineLineIntersectXY(box1[2],box2[1]):
        #     return True
        # if lineLineIntersectXY(box1[2],box2[3]):
        #     return True

        # if lineLineIntersectXY(box2[0],box1[1]):
        #     return True
        # if lineLineIntersectXY(box2[0],box1[3]):
        #     return True
        # if lineLineIntersectXY(box2[2],box1[1]):
        #     return True
        # if lineLineIntersectXY(box2[2],box1[3]):
        #     return True
        
        # for l1 in box1:
        #     for l2 in box2:
        #         if lineLineIntersectXY(l1,l2):
        #             return True
        return False
        
    # do projeted box line intersection tests
    return (int2D(bbx1,bbx2,'XY') and
            (not dim3 or int2D(bbx1,bbx2,'YZ')) and
            (not dim3 or int2D(bbx1,bbx2,'XZ')))
    
    
      
def boxoverlap(bbx1,bbx2,dim3=True):

    """determine if two bounding boxes overlap"""
    minx = bbx1[0][0]
    miny = bbx1[0][1]
    minz = bbx1[0][2]
    
    len1 = bbx1[1][0] - bbx1[0][0] # length
    wid1 = bbx1[1][1] - bbx1[0][1] # width
    hei1 = bbx1[1][2] - bbx1[0][2] # height

    p0=bbx1[0]
    p1=add(p0,point(len1,0,0))
    p2=add(p1,point(0,wid1,0))
    p3=add(p0,point(0,wid1,0))

    points = [p0,p1,p2,p3]
    
    if not dim3:
        for p in points:
            if isinsidebbox2D(bbx2,p):
                return True
        return False

    p4=add(p0,point(0,0,hei1))
    p5=add(p4,point(len1,0,0))
    p6=bbx1[1]
    p7=add(p4,point(0,wid1,0))

    points += [p4,p5,p6,p7]
    #print("boxoverlap points; ",vstr(points))
    #print("boxoverlap bbx2: ",vstr(bbx2))
    for p in points:
        if isinsidebbox(bbx2,p):
            return True
    return False
    
def bbox2oct(bbx,refbox,center):
    """
    Utility function to take a bounding box representation and assign
    it to zero or more octants.

    box2oct(bbx,refbox,center)

    bbx: 3D bounding box to assign
    refbox: reference 3D bounding box
    center: center point for purposes of assignment
      
    returns (potentially empty) list of octants, numbered 0 to 7
    """
    # print("bbox2oct :: bbx: ",vstr(bbx),"  refbox: ",vstr(refbox),"  center: ",vstr(center))
    rlist = []
    if not boxoverlap2(refbox,bbx):
        return rlist           # no overlap

    if bbx[0][2] < center[2]:
        rlist += bbox2quad(bbx,refbox,center)

    if bbx[1][2] >= center[2]:
        rlist += list(map(lambda x: x+4, bbox2quad(bbx,refbox,center)))

    return rlist
    
def bbox2quad(bbx,refbox,center):
    """Utility Function to take a bounding box representation and assign
       it to zero or more quads

    box2quad(bbx,refbox,center)

    bbx: 2D bounding box to assign
    refbox: reference 2D bounding box
    center: center point for purposes of assignment

    returns (potentially empty) list of quadrants, numbered 0 to 3
    """
    rlist = []
    if not boxoverlap2(refbox,bbx,dim3=False):
        return rlist           # no overlap

    if bbx[0][0] < center[0]:
        if bbx[0][1] < center[1]:
            rlist.append(2)
        if bbx[1][1] >= center[1]:
            rlist.append(1)
    if bbx[1][0] >= center[0]:
        if bbx[0][1] < center[1]:
            rlist.append(3)
        if bbx[1][1] >= center[1]:
            rlist.append(0)
    return rlist
            

def box2boxes(bbox,center,n,type='centersplit',elm=[]):
    """Function to take a bounding box (2d or 3D) and return a quad- or
    octtree decomposition of the box based on the value of center
    point, the type of split, and (potentially) the list of bounding
    boxes to be divvied up.
    """
    def boxmid(box):
        p0 = box[0]
        p1 = box[1]
        return scale3(add(p0,p1),0.5)

    # print("box2boxes :: bbox: ",vstr(bbox),"  center: ",vstr(center))
    if not isinsidebbox(bbox,center):
        raise ValueError('center point does not lie inside the bounding box')
    if type != 'centersplit':
        raise NotImplementedError('we are only doing center splits for now')
    if not n in (4,8):
        raise ValueError('bad tree dimension')

    cx = center[0]
    cy = center[1]
    cz = center[2]
        
    maxx = bbox[1][0]
    maxy = bbox[1][1]
    maxz = bbox[1][2]
    minx = bbox[0][0]
    miny = bbox[0][1]
    minz = bbox[0][2]

    if (maxx-minx < epsilon  or
        maxy-miny < epsilon or
        (n == 8 and maxz-minz < epsilon)) :
        raise ValueError('zero dimension box')
        
    if n == 4:
        z0 = 0
        z1 = 0
    else:
        z0 = minz
        z1 = cz
        
    box1 = [point(cx,cy,z0),
            point(maxx,maxy,z1)]
    box2 = [point(minx,cy,z0),
            point(cx,maxy,z1)]
    box3 = [point(minx,miny,z0),
            point(cx,cy,z1)]
    box4 = [point(cx,miny,z0),
            point(maxx,cy,z1)]

    r1 = [ [box1, boxmid(box1)],
           [box2, boxmid(box2)],
           [box3, boxmid(box3)],
           [box4, boxmid(box4)] ]

    if n == 4:
        return r1
    else:
        z0 = cz
        z1 = maxz
        
    box5 = [point(cx,cy,z0),
            point(maxx,maxy,z1)]
    box6 = [point(minx,cy,z0),
            point(cx,maxy,z1)]
    box7 = [point(minx,miny,z0),
            point(cx,cy,z1)]
    box8 = [point(cx,miny,z0),
            point(maxx,cy,z1)]

    r2 = [ [box5, boxmid(box5)],
           [box6, boxmid(box6)],
           [box7, boxmid(box7)],
           [box8, boxmid(box8)] ]

    return r1 + r2

def bboxdim(box):
    """ return length, width, and height of a bounding box"""

    length = box[1][0] - box[0][0]
    width = box[1][1] - box[0][1]
    height = box[1][2] - box[0][2]
    return [length, width, height]

def untag(tglist):
    """remove the (optional) tags that can be associated with elements in
    a geometry list."""
    glist = []
    for e in tglist:
        if isinstance(e,tuple):
            glist += [ e[0] ]
        else:
            glist += [ e ]
    return glist

class NTree():

    """Generalized n-tree representation for yapCAD geometry"""

    def __init__(self,n=8,geom=None,center=None,
                 mindim=None,maxdepth=None):

        if not n in [4,8]:
            raise ValueError('only quad- or octtrees supported')

        self.__n = n
        self.__depth = 0
        if mindim:
            if isinstance(mindim,(int,float)) and mindim > epsilon:
                self.__mindim = mindim
            else:
                raise ValueError('bad mindim value: '+str(mindim))
        else:
            self.__mindim = 1.0 # minimum dimension for tree element

        if maxdepth:
            if isinstance(maxdepth,int) and maxdepth > 0:
                self.__maxdepth = maxdepth
            else:
                raise ValueError('bad max depth value: '+str(maxdepth))
        elif self.__n == 4:
            self.__maxdepth = 8
        else: # n=8
            self.__maxdepth = 7

        self.__geom=[]
        if geom:
            ugeom = untag(geom)
            if isgeomlist(ugeom):
                self.__bbox= bbox(ugeom)
                self.__geom= geom
            else:
                raise ValueError('geom must be valid geometry list, or None')

        self.__center = None
        if center:
            if not ispoint(center):
                raise ValueError('center must be a valid point')
            self.__center = center
        else:
            if self.__geom != []:
                self.updateCenter()
                
        self.__tree = []
        self.__update=True

    def __repr__(self):
        return 'NTree(n={},depth={},mindim={},\n  geom={},\n  tree={})'.format(self.__n,self.__depth,self.__mindim,vstr(self.__geom),vstr(self.__tree))

    def addElement(self,element,tag=None):
        """ add a geometry element to the collection, don't update
        the tree -- yet """
        if not isgeomlist(element):
            raise ValueError('bad element passed to addElement')
        if not tag:
            gl = [ element ]
        else:
            gl = [ (element,tag) ]
        self.__geom += gl
        
        self.__update=True
        

    def updateCenter(self,center=None):
        """specify or compute new geometric center (or split poit) for tree,
        flag tree for rebuilding.

        """
        if self.__geom == []: #nothing to do
            return

        if not center:
            self.__center = scale3(add(self.__bbox[0],
                                       self.__bbox[1]),0.5)
        else:
            if ispoint(center):
                self.__center = center
            else:
                raise ValueError('bad center point passed to updateCenter')
        self.__update = True
        
        
    def updateTree(self):
        """
        build the tree from the current contents of the self.__geom list
        """
        if not self.__update or self.__geom == []:
            # nothing to do
            return
        self.__update = False

        # striptags for bbox
        ugeom = untag(self.__geom)
        self.__bbox = bbox(ugeom)
        if not self.__center:
            self.updateCenter()

        bxlist = list(map(lambda x: bbox(x), ugeom))
        bxidxlist = []
        for i in range(len(bxlist)):
            bxidxlist.append([i,bxlist[i]])
            
        self.__elem_idx_bbox = bxidxlist
        
        if not self.__center:
            self.updateCenter()

        # recursively build the tree
        def recurse(box,center,elements,depth=0):
            # print("recurse :: box: ",vstr(box),"  center: ",vstr(center))
            bbdim = bboxdim(box)
            if depth > self.__depth:
                self.__depth = depth
                
            if elements==[] or elements==None:
                return []
            elif (len(elements) <= self.__n or
                  depth > self.__maxdepth or
                  bbdim[0] < self.__mindim or
                  bbdim[1] < self.__mindim or
                  (self.__n == 8 and bbdim[2] < self.__mindim)):

                return [ 'e', box, center ] + list(map(lambda x:
                                                        x[0], elements))
            else:
                boxlist = ['b'] + box2boxes(box,center,self.__n)

                # print("boxlist: ",vstr(boxlist))
                # print("elements: ",vstr(elements))
                for e in elements:
                    func = None
                    box = e[1]
                    if self.__n == 8:
                        func = bbox2oct
                    else:
                        func = bbox2quad
                    ind = func(box,box,center)
                    # print("ind: ",ind,"  box: ",box,"  box: ",box,"  center: ",center)
                    #print ("e[0]: ",e[0], "  e[1]: ",vstr(e[1]),"  ind: ",ind)
                    # print ("box: ",vstr(box))
                    for j in ind:
                        boxlist[j+1].append(e)
                        
                for i in range(1,len(boxlist)):
                    boxlist[i] = recurse(boxlist[i][0],boxlist[i][1],
                                         boxlist[i][2:],depth+1)
                return boxlist
                
        self.__tree = recurse(self.__bbox, self.__center,
                              bxidxlist)
        # print("bxidxlist: ",vstr(bxidxlist))
        # print("self.__tree",self.__tree)

        return

    @property
    def depth(self):
        return self.__depth

    @depth.setter
    def depth(self,n):
        raise ValueError("can't set tree depth")

    @property
    def maxdepth(self):
        return self.__maxdepth

    @maxdepth.setter
    def maxdepth(self,d):
        if not isinstance(d,int) or d < 1:
            raise ValueError('bad maxdepth value: '+str(d))
        else:
            self.__maxdepth = d
    
    @property
    def mindim(self):
        return self.__mindim

    @mindim.setter
    def mindim(self,d):
        if not isinstance(d,(float,int)) or d < epsilon:
            raise ValueError('bad mindim value: '+str(d))
        else:
            self.__mindim = d
    
    def getElements(self,bbox):
        """return a list of geometry elements with bounding boxes that
        overalp the provided bounding box, or the empty list if none.

        """
        if self.__update:
            self.updateTree()

        bxidxlist = self.__elem_idx_bbox

        self.mxd = 0
        
        def recurse(subtree,depth=0):
            indices = []
            if depth > self.mxd:
                self.mxd = depth

            if subtree == None or subtree == []:
                return []
            elif subtree[0] == 'e':
                if boxoverlap2(subtree[1],bbox,dim3 = (self.__n==8)):
                    for ind in subtree[3:]:
                        box = bxidxlist[ind][1]
                        if boxoverlap2(bbox,box,dim3 = (self.__n==8)): 
                            indices.append(ind)
            else:
                for boxlist in subtree[1:]:
                    if len(subtree) == 1:
                        print('subtree ',subtree)
                    if len(boxlist) > 0:
                        indices += recurse(boxlist,depth+1)
            return indices

        idx = set(recurse(self.__tree))
                
        #print("unique indices: ",idx)

        elements = list(map(lambda x: self.__geom[x], list(idx)))

        return elements



# -*- coding: utf-8 -*-
try:  # Python >= 3.8
    from importlib.metadata import PackageNotFoundError, version
except ModuleNotFoundError:  # pragma: no cover - for Python < 3.8
    from importlib_metadata import PackageNotFoundError, version


try:
    __version__ = version("yapCAD")
except PackageNotFoundError:  # pragma: no cover - handled when package not installed
    __version__ = "unknown"
